"""
Script to obtain fractional abundances from Aurora's reading of ADAS data.
Note that this is a calculation from atomic physics only, i.e. no transport is applied.

It is recommended to run this in IPython.
"""
import numpy as np
import matplotlib.pyplot as plt

plt.ion()
import sys
import os
import time
from scipy.interpolate import interp1d
from matplotlib import cm

# Make sure that package home is added to sys.path
sys.path.append("../")
import aurora

# pass any argument via the command line to show plots
plot = len(sys.argv) > 1

ion = "Al"

# create some arbitrary kinetic profiles for this example
# parameterization f=(f_center-f_edge)*(1-rho**alpha1)**alpha2 + f_edge
T_core = 5e3  # eV
T_edge = 100  # eV
T_alpha1 = 2.0
T_alpha2 = 1.5
n_core = 1e14  # cm^-3
n_edge = 0.4e14  # cm^-3
n_alpha1 = 2
n_alpha2 = 0.5

# now define kinetic profiles
rhop = np.linspace(0, 1, 100)
ne_cm3 = (n_core - n_edge) * (1 - rhop ** n_alpha1) ** n_alpha2 + n_edge
Te_eV = (T_core - T_edge) * (1 - rhop ** T_alpha1) ** T_alpha2 + T_edge

# get charge state distributions from ionization equilibrium for Ca
atom_data = aurora.atomic.get_atom_data(ion, ["scd", "acd", "ccd"])

# get fractional abundances on ne (cm^-3) and Te (eV) grid
_Te, fz = aurora.atomic.get_frac_abundances(
    atom_data, ne_cm3, Te_eV, rho=rhop, plot=plot
)

# include effect of CX with a given (here, arbitrary) density of neutrals
n0_by_ne = (
    1e-2 * np.exp(rhop ** 5 - 1)
) ** 2  # arbitrary, exponentially decreasing from LCFS
_Te, fz = aurora.atomic.get_frac_abundances(
    atom_data,
    ne_cm3,
    Te_eV,
    n0_by_ne,
    rho=rhop,
    plot=plot,
    ax=plt.gca() if plot else None,
)

# compare to fractial abundances obtained with ne*tau=1e19 m^-3.s
# Use the get_atomic_relax_time function to use a finite ne*tau value
_Te, fz, rate_coeffs = aurora.get_atomic_relax_time(
    atom_data, ne_cm3, Te_eV, ne_tau=1e19, plot=False
)

if plot:
    # overplot using cubic interpolation for smoother visualization:
    plt.gca().set_prop_cycle("color", cm.plasma(np.linspace(0, 1, fz.shape[1])))
    x_fine = np.linspace(np.min(rhop), np.max(rhop), 10000)
    fz_i = interp1d(rhop, fz, kind="cubic", axis=0)(x_fine)
    plt.gca().plot(x_fine, fz_i, ls="--")

# plot atomic relaxation time over Te (no transport) at a single value of ne
_Te, fz, rate_coeffs = aurora.get_atomic_relax_time(
    atom_data, np.ones(100) * 1e14, np.logspace(1, 4, 100), ne_tau=np.inf, plot=plot
)



"""Functions to provide default ADAS files for Aurora modelling, including capabilities to fetch
these files remotely from the OPEN-ADAS website.
"""
# MIT License
#
# Copyright (c) 2021 Francesco Sciortino
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import shutil, os
import requests

# location of the "adas_data" directory relative to this script:
adas_data_dir = (
    os.path.dirname(os.path.realpath(__file__)) + os.sep + "adas_data" + os.sep
)


def get_adas_file_loc(filename, filetype="adf11"):
    """Find location of requested atomic data file for the indicated ion. 
    The search proceeds with the following attempts, in this order:

    #. If the file is available in Aurora/adas_data/*filetype*, with filetype given by the user, 
       always use this data.

    #. If the input filename is actually a full path to the file on the local system, use this file 
       and copy it to Aurora/aurora/adas_data/actual_filename, where actual_filename is the file
       name rather than the full path. 

    #. If the environmental variable "AURORA_ADAS_DIR" is defined, attempt to find the file there, 
       under adf11/ and adf15/ directories. AURORA_ADAS_DIR may for example be the path to a central
       repository of ADAS files for Aurora on a cluster where not everyone may have write-permissions.
       For this option, files are not copied at all.

    #. Attempt to fetch the file remotely via open.adas.ac.uk and save it in 
       Aurora/aurora/adas_data/*filetype*/. 

    Parameters
    ----------
    filename : str
        Name of the ADAS file of interest, e.g. 'plt89_ar.dat'.
    filetype : str
        ADAS file type. Currently allowed: 'adf11' or 'adf15'
    
    Returns
    -------
    file_loc : str
        Full path to the requested file. 
    """

    def fetch_file(filename, filetype, loc):
        if not os.path.isdir(os.path.dirname(os.path.dirname(loc))):
            # make sure that aurora/adas_data/adf** directory exists, so that we can store data files in it
            os.makedirs(os.path.dirname(os.path.dirname(loc)))
        if not os.path.isdir(os.path.dirname(loc)):
            # make sure that aurora/adas_data/adf** directory exists, so that we can store data files in it
            os.makedirs(os.path.dirname(loc))

        if filetype == "adf11":
            fetch_adf11_file(filename, loc)
        elif filetype == "adf15":
            fetch_adf15_file(filename, loc)
        else:
            raise ValueError(
                "ADAS file type/format not recognized. Could not find it or download it automatically!"
            )

    if os.path.exists(adas_data_dir + filetype + os.sep + filename):
        # file is available in adas_data:
        return adas_data_dir + filetype + os.sep + filename

    elif os.path.exists(filename):
        # user gave a complete filepath. Don't copy the file from the original location
        return filename

    elif "AURORA_ADAS_DIR" in os.environ:
        loc = os.environ["AURORA_ADAS_DIR"] + os.sep + filetype + os.sep + filename
        if os.path.exists(loc):
            pass
        else:
            # File could not be found. Download it and save it in AURORA_ADAS_DIR:
            fetch_file(filename, filetype, loc=loc)
        return loc

    else:
        # File could not be found. Download it and save it in adas_data:
        if not os.path.isdir(adas_data_dir):
            # make sure that aurora/adas_data exists
            os.makedirs(adas_data_dir)

        loc = adas_data_dir + filetype + os.sep + filename
        fetch_file(filename, filetype, loc=loc)
        return loc


def fetch_adf11_file(filename, loc):
    """Download ADF11 file from the OPEN-ADAS website and store it in the 'adas_data/adf11'
    directory. 

    Parameters
    ----------
    filename : str
        Name of ADF11 file to be downloaded, e.g. 'plt89_ar.dat'.
    loc : str
        Location to save fetched ADF11 in.
    """
    url = "https://open.adas.ac.uk/download/adf11/"
    str1 = filename.split("_")[0]

    r = requests.get(url + str1 + "/" + filename)

    if len(r.text) < 1000:
        # OPEN-ADAS reports short URL error text rather than an error code
        raise ValueError(f"Could not fetch file {filename} from ADAS!")

    with open(loc, "wb") as f:
        f.write(r.content)


def fetch_adf15_file(filename, loc):
    """Download ADF15 file from the OPEN-ADAS website and store it in the 'adas_data/adf15'
    directory. 

    Parameters
    ----------
    filename : str
        Name of ADF15 file to be downloaded, e.g. 'pec96#c_pju#c2.dat'.
    loc : str
        Location to save fetched ADF15 file in.
    """
    url = "https://open.adas.ac.uk/download/adf15/"

    if filename.startswith("pec"):
        # more standard format, the following patterns should work fine:
        num = filename[3:5]
        spec = filename.split("#")[1].split("_")[0]
        filename_mod = filename.replace("#", "][").replace("_", f"/pec{num}][{spec}_")
    elif filename.startswith("transport"):
        # different link format for files with "transport" name:
        filename_mod = "transport/" + filename.replace("#", "][")
    else:
        # patterns may be different, attempt simple guess:
        filename_mod = filename.split("_")[0] + "/" + filename.replace("#", "][")

    r = requests.get(url + "/" + filename_mod)

    if len(r.text) < 1000:
        # OPEN-ADAS reports short URL error text rather than an error code
        raise ValueError(f"Could not fetch file {filename_mod} from ADAS!")

    with open(loc, "wb") as f:
        f.write(r.content)


def adas_files_dict():
    """Selections for ADAS files for Aurora runs and radiation calculations.
    This function can be called to fetch a set of default files, which can then be modified (e.g. to 
    use a new file for a specific SXR filter) before running a calculation. 

    Returns
    -------
    files : dict
        Dictionary with keys equal to the atomic symbols of many of the most common ions of
        interest in fusion research. For each ion, a sub-dictionary contains recommended file 
        names for the relevant ADAS data types. Not all files types are available for all ions. 
        Files types are usually a subset of 
        'acd','scd','prb','plt','ccd','prc','pls','prs','fis','brs','pbs',prc'
        Refer to :py:func:`~aurora.atomic.get_adas_file_types` for a description of the meaning of 
        each file.
    """

    files = {}
    files["H"] = {}  # 1
    files["H"]["acd"] = "acd12_h.dat"  # "acd96_h.dat"
    files["H"]["scd"] = "scd12_h.dat"  # "scd96_h.dat"
    files["H"]["prb"] = "prb12_h.dat"  # "prb96_h.dat"
    files["H"]["plt"] = "plt12_h.dat"  # "plt96_h.dat"
    files["H"]["ccd"] = "ccd96_h.dat"
    files["H"]["prc"] = "prc96_h.dat"
    files["H"]["pls"] = "pls_H_14.dat"
    files["H"]["prs"] = "prs_H_14.dat"
    files["H"]["fis"] = "sxrfil14.dat"
    files["H"]["brs"] = "brs05360.dat"
    files["H"]["pbs"] = "pbsx7_h.dat"
    files["H"]["prc"] = "prc89_h.dat"
    files["He"] = {}  # 2
    files["He"]["acd"] = "acd96_he.dat"
    files["He"]["scd"] = "scd96_he.dat"
    files["He"]["prb"] = "prb96_he.dat"
    files["He"]["plt"] = "plt96_he.dat"
    files["He"]["ccd"] = "ccd96_he.dat"
    files["He"]["pls"] = "pls_He_14.dat"
    files["He"]["prs"] = "prs_He_14.dat"
    files["He"]["fis"] = "sxrfil14.dat"
    files["He"]["brs"] = "brs05360.dat"
    files["He"]["pbs"] = "pbsx5_he.dat"
    files["He"]["prc"] = "prc96_he.dat"
    files["Li"] = {}  # 3
    files["Li"]["acd"] = "acd96_li.dat"
    files["Li"]["scd"] = "scd96_li.dat"
    files["Li"]["ccd"] = "ccd89_li.dat"
    files["Li"]["prb"] = "prb96_li.dat"
    files["Li"]["plt"] = "plt96_li.dat"
    files["Li"]["prc"] = "prc89_li.dat"
    files["Li"]["pls"] = "pls89_li.dat"
    files["Li"]["pbs"] = ""
    files["Li"]["prc"] = "prc89_li.dat"
    files["Be"] = {}  # 4
    files["Be"]["acd"] = "acd96_be.dat"
    files["Be"]["scd"] = "scd96_be.dat"
    files["Be"]["prb"] = "prb96_be.dat"
    files["Be"]["plt"] = "plt96_be.dat"
    files["Be"]["ccd"] = "ccd89_be.dat"
    files["Be"]["prc"] = "prc89_be.dat"
    files["Be"]["pls"] = "plsx5_be.dat"
    files["Be"]["prs"] = "prsx5_be.dat"
    files["Be"]["pbs"] = "pbsx5_be.dat"
    files["Be"]["prc"] = "prc89_be.dat"
    files["B"] = {}  # 5
    files["B"]["acd"] = "acd89_b.dat"
    files["B"]["scd"] = "scd89_b.dat"
    files["B"]["ccd"] = "ccd89_b.dat"
    files["B"]["prb"] = "prb89_b.dat"
    files["B"]["plt"] = "plt89_b.dat"
    files["B"]["prc"] = "prc89_b.dat"
    files["B"]["pls"] = "plsx5_b.dat"
    files["B"]["prs"] = "prsx5_b.dat"
    files["B"]["pbs"] = "pbsx5_b.dat"
    files["B"]["prc"] = "prc89_b.dat"
    files["C"] = {}  # 6
    files["C"]["acd"] = "acd96_c.dat"
    files["C"]["scd"] = "scd96_c.dat"
    files["C"]["prb"] = "prb96_c.dat"
    files["C"]["plt"] = "plt96_c.dat"
    files["C"]["ccd"] = "ccd96_c.dat"
    files["C"]["pls"] = "pls_C_14.dat"
    files["C"]["prs"] = "prs_C_14.dat"
    files["C"]["fis"] = "sxrfil14.dat"
    files["C"]["brs"] = "brs05360.dat"
    files["C"]["pbs"] = "pbsx5_c.dat"
    files["C"]["prc"] = "prc96_c.dat"
    files["N"] = {}  # 7
    files["N"]["acd"] = "acd96_n.dat"
    files["N"]["scd"] = "scd96_n.dat"
    files["N"]["ccd"] = "ccd89_n.dat"
    files["N"]["prb"] = "prb96_n.dat"
    files["N"]["plt"] = "plt96_n.dat"
    files["N"]["pls"] = "plsx8_n.dat"
    files["N"]["prs"] = "prsx8_n.dat"
    files["N"]["fis"] = "sxrfilD1.dat"
    files["N"]["brs"] = "brs05360.dat"
    files["N"]["ccd"] = "ccd96_n.dat"
    files["N"]["pbs"] = "pbsx5_n.dat"
    files["N"]["prc"] = "prc96_n.dat"
    files["O"] = {}  # 8
    files["O"]["acd"] = "acd96_o.dat"
    files["O"]["scd"] = "scd96_o.dat"
    files["O"]["ccd"] = "ccd89_o.dat"
    files["O"]["prb"] = "prb96_o.dat"
    files["O"]["plt"] = "plt96_o.dat"
    files["O"]["pls"] = "plsx5_o.dat"
    files["O"]["prs"] = "prsx5_o.dat"
    files["O"]["pbs"] = "pbsx5_o.dat"
    files["O"]["prc"] = "prc89_o.dat"
    files["F"] = {}  # 9
    files["F"][
        "acd"
    ] = "acd89_f.dat"  # sub with Puetterich data when available from OPEN-ADAS
    files["F"][
        "scd"
    ] = "scd89_f.dat"  # sub with Puetterich data when available from OPEN-ADAS
    files["F"]["ccd"] = "ccd89_f.dat"
    files["F"]["prb"] = "prb89_f.dat"
    files["F"]["plt"] = "plt89_f.dat"
    files["F"]["fis"] = "sxrfil14.dat"
    files["F"]["brs"] = "brs05360.dat"
    files["F"]["pls"] = "pls_F_14.dat"
    files["F"]["prs"] = "prs_F_14.dat"
    files["F"]["pbs"] = "pbsx5_f.dat"
    files["F"]["prc"] = "prc89_f.dat"
    files["Ne"] = {}  # 10
    files["Ne"]["acd"] = "acd96_ne.dat"
    files["Ne"]["scd"] = "scd96_ne.dat"
    files["Ne"]["prb"] = "prb96_ne.dat"
    files["Ne"]["plt"] = "plt96_ne.dat"
    files["Ne"]["ccd"] = "ccd89_ne.dat"
    files["Ne"]["pls"] = "plsx8_ne.dat"
    files["Ne"]["prs"] = "prsx8_ne.dat"
    files["Ne"]["fis"] = "sxrfilD1.dat"
    files["Ne"]["brs"] = "brs05360.dat"
    files["Ne"]["pbs"] = "pbsx5_ne.dat"
    files["Ne"]["prc"] = "prc89_ne.dat"
    files["Na"] = {}  # 10
    files["Na"]["acd"] = "acd85_ne.dat"
    files["Na"]["scd"] = "scd85_ne.dat"
    # files["Na"]['prb'] = ""
    # files["Na"]['plt'] = ""
    # files["Na"]['ccd'] = ""
    # files["Na"]['pls'] = ""
    # files["Na"]['prs'] = ""
    # files["Na"]['fis'] = ""
    # files["Na"]['brs'] = ""
    # files["Na"]["pbs"] = ""
    # files["Na"]['prc'] = ""
    files["Mg"] = {}  # 12
    files["Mg"]["acd"] = "acd85_mg.dat"
    files["Mg"]["scd"] = "scd85_mg.dat"
    # files["Mg"]['prb'] = ""
    # files["Mg"]['plt'] = ""
    # files["Mg"]['ccd'] = ""
    # files["Mg"]['pls'] = ""
    # files["Mg"]['prs'] = ""
    # files["Mg"]['fis'] = ""
    # files["Mg"]['brs'] = ""
    # files["Mg"]["pbs"] = ""
    # files["Mg"]['prc'] = ""
    files["Al"] = {}  # 13
    files["Al"]["acd"] = "acd89_al.dat"
    files["Al"]["scd"] = "scd89_al.dat"
    files["Al"]["prb"] = "prb89_al.dat"
    files["Al"]["plt"] = "plt89_al.dat"
    files["Al"]["ccd"] = "ccd89_al.dat"
    files["Al"]["pls"] = "pls_Al_14.dat"
    files["Al"]["prs"] = "prs_Al_14.dat"
    files["Al"]["fis"] = "sxrfil14.dat"
    files["Al"]["brs"] = "brs05360.dat"
    files["Al"]["pbs"] = "pbsx5_al.dat"
    files["Al"]["prc"] = "prc89_al.dat"
    files["Si"] = {}  # 14
    files["Si"]["acd"] = "acd96_si.dat"
    files["Si"]["scd"] = "scd96_si.dat"
    files["Si"]["prb"] = "prb96_si.dat"
    files["Si"]["plt"] = "plt96_si.dat"
    files["Si"]["pls"] = "pls_Si_14.dat"
    files["Si"]["prs"] = "prs_Si_14.dat"
    files["Si"]["fis"] = "sxrfil14.dat"
    files["Si"]["brs"] = "brs05360.dat"
    files["Si"]["ccd"] = "ccd89_si.dat"
    files["Si"]["pbs"] = "pbsx5_si.dat"
    files["Si"]["prc"] = "prc89_si.dat"
    files["S"] = {}  # 16
    files["S"]["acd"] = "acd89_s.dat"
    files["S"]["scd"] = "scd89_s.dat"
    files["S"]["prb"] = "prb89_s.dat"
    files["S"]["plt"] = "plt89_s.dat"
    files["S"]["pls"] = "pls_S_14.dat"
    files["S"]["prs"] = "prs_S_14.dat"
    files["S"]["fis"] = "sxrfil14.dat"
    files["S"]["brs"] = "brs05360.dat"
    files["S"]["ccd"] = "ccd89_s.dat"
    files["S"]["pbs"] = "pbsx5_s.dat"
    files["S"]["prc"] = "prc89_s.dat"
    files["Cl"] = {}  # 17
    files["Cl"]["acd"] = "acd89_cl.dat"
    files["Cl"]["scd"] = "scd89_cl.dat"
    files["Cl"]["prb"] = "prb89_cl.dat"
    files["Cl"]["plt"] = "plt89_cl.dat"
    files["Cl"]["pls"] = "pls_Cl_14.dat"
    files["Cl"]["prs"] = "prs_Cl_14.dat"
    files["Cl"]["fis"] = "sxrfil14.dat"
    files["Cl"]["brs"] = "brs05360.dat"
    files["Cl"]["ccd"] = "ccd89_cl.dat"
    files["Cl"]["pbs"] = "pbsx5_cl.dat"
    files["Cl"]["prc"] = "prc89_cl.dat"
    files["Ar"] = {}  # 18
    files["Ar"]["acd"] = "acd89_ar.dat"
    files["Ar"]["scd"] = "scd89_ar.dat"
    files["Ar"]["prb"] = "prb89_ar.dat"
    files["Ar"]["plt"] = "plt41_ar.dat"  # Henderson data
    files["Ar"]["ccd"] = "ccd89_ar.dat"
    files["Ar"]["prc"] = "prc89_ar.dat"
    files["Ar"]["pls"] = "pls_Ar_14.dat"
    files["Ar"]["prs"] = "prs_Ar_14.dat"
    files["Ar"]["fis"] = "sxrfil14.dat"
    files["Ar"]["brs"] = "brs05360.dat"
    files["Ar"]["pbs"] = "pbsx5_ar.dat"
    files["Ar"]["prc"] = "prc89_ar.dat"
    files["Ca"] = {}  # 20
    files["Ca"]["acd"] = "acd85_ca.dat"
    files["Ca"]["scd"] = "scd85_ca.dat"
    files["Ca"][
        "ccd"
    ] = "ccd89_w.dat"  # Ca CCD file not available, use first 20 ion stages using Foster scaling
    files["Ca"]["prb"] = "prb85_ca.dat"  # not public on OPEN-ADAS, must request
    files["Ca"]["plt"] = "plt85_ca.dat"  # not public on OPEN-ADAS, must request
    files["Ca"]["pls"] = "pls_Ca_14.dat"
    files["Ca"]["prs"] = "prs_Ca_14.dat"
    files["Ca"]["fis"] = "sxrfil14.dat"
    files["Ca"]["brs"] = "brs05360.dat"
    files["Ca"]["pbs"] = ""
    files["Ca"]["prc"] = ""
    files["Fe"] = {}  # 26
    files["Fe"]["acd"] = "acd89_fe.dat"
    files["Fe"]["scd"] = "scd89_fe.dat"
    files["Fe"]["prb"] = "prb89_fe.dat"
    files["Fe"]["plt"] = "plt41_fe.dat"  # Henderson data
    files["Fe"]["pls"] = "pls_Fe_14.dat"
    files["Fe"]["prs"] = "prs_Fe_14.dat"
    files["Fe"]["fis"] = "sxrfil14.dat"
    files["Fe"]["brs"] = "brs05360.dat"
    files["Fe"]["ccd"] = "ccd89_fe.dat"
    files["Fe"]["pbs"] = "pbsx5_fe.dat"
    files["Fe"]["prc"] = "prc89_fe.dat"
    files["Ni"] = {}  # 28
    files["Ni"]["acd"] = "acd85_ni.dat"
    files["Ni"]["scd"] = "scd85_ni.dat"
    files["Ni"]["prb"] = "prb89_ni.dat"
    files["Ni"]["plt"] = "plt89_ni.dat"
    files["Ni"]["pls"] = "pls_Ni_14.dat"
    files["Ni"]["prs"] = "prs_Ni_14.dat"
    files["Ni"]["fis"] = "sxrfil14.dat"
    files["Ni"]["brs"] = "brs05360.dat"
    files["Ni"]["ccd"] = "ccd89_ni.dat"
    files["Ni"]["pbs"] = "pbsx5_ni.dat"
    files["Ni"]["prc"] = "prc89_ni.dat"
    files["Kr"] = {}  # 36
    files["Kr"]["acd"] = "acd89_kr.dat"
    files["Kr"]["scd"] = "scd89_kr.dat"
    files["Kr"]["ccd"] = "ccd89_kr.dat"
    files["Kr"]["prb"] = "prb89_kr.dat"
    files["Kr"]["plt"] = "plt41_kr.dat"  # Henderson data
    files["Kr"]["pls"] = "plsx5_kr.dat"
    files["Kr"]["prs"] = "prsx5_kr.dat"
    files["Kr"]["pbs"] = "pbsx5_kr.dat"
    files["Kr"]["prc"] = "prc89_kr.dat"
    files["Mo"] = {}  # 42
    files["Mo"]["acd"] = "acd89_mo.dat"
    files["Mo"]["scd"] = "scd89_mo.dat"
    files["Mo"]["ccd"] = "ccd89_mo.dat"
    files["Mo"]["plt"] = "plt89_mo.dat"
    files["Mo"]["prb"] = "prb89_mo.dat"
    files["Mo"]["pls"] = "plsx5_mo.dat"
    files["Mo"]["prs"] = "prsx5_mo.dat"
    files["Mo"]["pbs"] = "pbsx5_mo.dat"
    files["Mo"]["prc"] = "prc89_mo.dat"
    files["Xe"] = {}  # 56
    files["Xe"]["acd"] = "acd89_xe.dat"
    files["Xe"]["scd"] = "scd89_xe.dat"
    files["Xe"]["ccd"] = "ccd89_xe.dat"
    files["Xe"]["plt"] = "plt41_xe.dat"  # Henderson data
    files["Xe"]["prb"] = "prb89_xe.dat"
    files["Xe"]["prs"] = "prsx5_xe.dat"
    files["Xe"]["pls"] = "prsx5_xe.dat"
    files["Xe"]["pbs"] = "pbsx5_xe.dat"
    files["Xe"]["prc"] = "prc89_xe.dat"
    files["W"] = {}  # 74
    files["W"]["acd"] = "acd89_w.dat"
    files["W"]["scd"] = "scd89_w.dat"
    files["W"]["prb"] = "prb89_w.dat"
    files["W"]["plt"] = "plt41_w.dat"
    files["W"]["fis"] = "sxrfil14.dat"
    files["W"]["brs"] = "brs05360.dat"
    files["W"]["pls"] = "pls_W_14.dat"
    files["W"]["prs"] = "prs_W_14.dat"
    files["W"]["ccd"] = "ccd89_w.dat"
    files["W"]["pbs"] = "pbsx5_w.dat"
    files["W"]["prc"] = "prc89_w.dat"

    return files



# MIT License
#
# Copyright (c) 2021 Francesco Sciortino
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np


def load_default_namelist():
    """ Load default namelist. 
    Users should modify and complement this for a successful forward-model run.
    """

    namelist = {
        "imp": "Ca",
        "main_element": "D",  # background
        # --------------------
        "source_rate": 1e21,
        "source_type": "const",
        # explicit source:
        "explicit_source_vals": None,  # provide 2D array on explicit_source_time and explicit_source_rhop grids
        "explicit_source_time": None,
        "explicit_source_rhop": None,
        # other options to specify source:
        "source_width_in": 0.0,  # exponential ionization decay from wall boundary if widths both = 0
        "source_width_out": 0.0,
        "imp_source_energy_eV": 3.0,  # only needed if source_width_in=source_width_out=0.0
        "prompt_redep_flag": False,
        "source_file": None,  # required if source_type='file'
        "source_cm_out_lcfs": 1.0,  # source distance in cm from LCFS
        # LBO synthetic model, only used if source_type='synth_LBO'
        "LBO": {"n_particles": 1e18, "t_fall": 0.3, "t_rise": 0.05, "t_start": 0.01},
        # --------------------
        # time grid specifications
        "timing": {
            "dt_increase": np.array([1.005, 1.0]),
            "dt_start": np.array([1e-5, 0.001]),
            "steps_per_cycle": np.array([1, 1]),
            "times": np.array([0.0, 0.1]),
        },
        # --------------------
        # radial grid
        "bound_sep": 2.0,
        "lim_sep": 1.0,
        "clen_divertor": 17.0,
        "clen_limiter": 0.5,
        "dr_0": 0.3,
        "dr_1": 0.05,
        "K": 6.0,
        "SOL_decay": 0.05,
        # -------------------
        "saw_model": {
            "saw_flag": False,
            "rmix": 1000.0,
            "times": [1.0,],
            "crash_width": 1.0,
        },
        # --------------------
        # edge/recycling
        "recycling_flag": False,
        "wall_recycling": 0.0,
        "tau_div_SOL_ms": 50.0,  # ms
        "tau_pump_ms": 500.0,  # ms
        "tau_rcl_ret_ms": 50.0,  # ms
        "SOL_mach": 0.1,
        # --------------------
        # kinetic profiles
        "kin_profs": {
            "ne": {"fun": "interpa", "times": [1.0]},
            "Te": {"fun": "interp", "times": [1.0], "decay": [1.0]},
            "Ti": {"fun": "interp", "times": [1.0], "decay": [1.0]},
            "n0": {"fun": "interpa", "times": [1.0]},
        },
        # and NBI CXR rates
        "nbi_cxr": {"rhop": None, "vals": None},  # time indpt profiles; (rhop,nZ)
        # ----------------
        # charge exchange flags
        "cxr_flag": False,
        "nbi_cxr_flag": False,
        # -----------------
        # misc
        "device": "CMOD",
        "shot": 99999,
        "time": 1250,  # in ms, for equilibrium
        # atomic data (ADF11)
        "acd": None,  # use default
        "scd": None,  # use default
        "ccd": None,  # use default
        "superstages": [],
    }
    return namelist


if __name__ == "__main__":
    # print default namelist for documentation

    nml = load_default_namelist()
    for key in nml:
        if isinstance(nml[key], dict):
            for key2 in nml[key]:
                print(f'   * - `{key}["{key2}"]`')
                print(f"     - {nml[key][key2]}")
                print("     -")
        else:
            print(f"   * - `{key}`")
            print(f"     - {nml[key]}")
            print("     -")



"""
Script to test simple functionality of radiation models.

It is recommended to run this in IPython.
"""

import numpy as np
import matplotlib.pyplot as plt

plt.ion()
from omfit_classes import omfit_eqdsk
import sys, os
from scipy.interpolate import interp1d

# Make sure that package home is added to sys.path
sys.path.append("../")
import aurora

try:  # pass any argument via the command line to show plots
    plot = len(sys.argv) > 1
except:
    plot = False

# Use gfile and statefile in local directory:
examples_dir = os.path.dirname(os.path.abspath(__file__))
geqdsk = omfit_eqdsk.OMFITgeqdsk(examples_dir + "/example.gfile")

# save kinetic profiles on a rhop (sqrt of norm. pol. flux) grid
# parameterization f=(f_center-f_edge)*(1-rhop**alpha1)**alpha2 + f_edge
T_core = 5e3  # eV
T_edge = 100  # eV
T_alpha1 = 2.0
T_alpha2 = 1.5
n_core = 1e14  # cm^-3
n_edge = 0.4e14  # cm^-3
n_alpha1 = 2
n_alpha2 = 0.5

rhop = np.linspace(0, 1, 100)
ne_cm3 = (n_core - n_edge) * (1 - rhop ** n_alpha1) ** n_alpha2 + n_edge
Te_eV = (T_core - T_edge) * (1 - rhop ** T_alpha1) ** T_alpha2 + T_edge

# choice of ion and fractional abundance wrt electron density:
ion = "Ar"
frac = 0.005

# look at radiation profiles
res = aurora.radiation_model(ion, rhop, ne_cm3, Te_eV, geqdsk, frac=frac, plot=True)



"""
Script to demonstrate capabilities to load and post-process SOLPS results.
Note that SOLPS output is not distributed with Aurora; so, in order to run this script,
either you are able to access the default AUG SOLPS MDS+ server, or you need
to appropriately modify the script to point to your own SOLPS results.
"""
import numpy as np
import matplotlib.pyplot as plt

plt.ion()
from omfit_classes import omfit_eqdsk
import sys, os
import aurora

# if one wants to load a SOLPS case from MDS+ (defaults for AUG):
so = aurora.solps_case(solps_id=141349)

# alternatively, one may want to load SOLPS results from files on disk:
so2 = aurora.solps_case(
    b2fstate_path="/afs/ipp/home/s/sciof/SOLPS/141349/b2fstate",
    b2fgmtry_path="/afs/ipp/home/s/sciof/SOLPS/141349/b2fgmtry",
)

# plot some important fields
fig, axs = plt.subplots(1, 2, figsize=(10, 6), sharex=True)
ax = axs.flatten()
so.plot2d_b2(so.data("ne"), ax=ax[0], scale="log", label=r"$n_e$ [$m^{-3}$]")
so.plot2d_b2(so.data("te"), ax=ax[1], scale="linear", label=r"$T_e$ [eV]")


if hasattr(so, "fort46") and hasattr(so, "fort44"):
    # if EIRENE data files (e.g. fort.44, .46, etc.) are available,
    # one can plot EIRENE results on the original EIRENE grid.
    # SOLPS results also include EIRENE outputs on B2 grid
    fig, axs = plt.subplots(1, 2, figsize=(10, 6), sharex=True)
    so.plot2d_eirene(
        so.fort46["pdena"][:, 0] * 1e6,
        scale="log",
        label=r"$n_n$ [$m^{-3}$]",
        ax=axs[0],
    )
    so.plot2d_b2(so.fort44["dab2"][:, :, 0].T, label=r"$n_n$ [$m^{-3}$]", ax=axs[1])
    plt.tight_layout()



"""Aurora functionality for edge neutral modeling. 
The ehr5 file from DEGAS2 is used. See https://w3.pppl.gov/degas2/ for details.
"""
# MIT License
#
# Copyright (c) 2021 Francesco Sciortino
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from matplotlib import cm
from scipy.optimize import curve_fit, least_squares
import numpy as np
import matplotlib.pyplot as plt

plt.ion()
import urllib
import shutil, os, copy
from scipy.constants import e, h, c as c_light, Rydberg
from scipy.interpolate import interp2d
import requests

from . import plot_tools
from . import radiation
from . import adas_files

if "AURORA_ADAS_DIR" in os.environ:
    # if user indicated a directory for atomic data, use that
    atomic_data_dir = os.environ["AURORA_ADAS_DIR"]
else:
    # location of the "adas_data" directory relative to this script:
    atomic_data_dir = (
        os.path.dirname(os.path.realpath(__file__)) + os.sep + "adas_data" + os.sep
    )


def download_ehr5_file():
    """Download the ehr5.dat file containing atomic data describing the multi-step ionization and 
    recombination of hydrogen. 

    See https://w3.pppl.gov/degas2/ for details. 
    """
    filename = "ehr5.dat"
    url = "https://w3.pppl.gov/degas2/ehr5.dat"
    r = requests.get(url)

    # write file to directory where ADAS data is also located
    with open(atomic_data_dir + "/ehr5.dat", "wb") as f:
        f.write(r.content)

    print("Successfully downloaded the DEGAS2 ehr5.dat file.")


class ehr5_file:
    """ Read ehr5.dat file from DEGAS2. 
    Returns a dictionary containing

    - Ionization rate Seff in :math:`cm^3 s^{-1}`
    - Recombination rate Reff in :math:`cm^3 s^{-1}`
    - Neutral electron losses :math:`E_{loss}^{(i)}` in :math:`erg s^{-1}`
    - Continuum electron losses :math:`E_{loss}^{(ii)}` in :math:`erg s^{-1}`
    - Neutral “n=2 / n=1”, :math:`N_2^{(i)}/N_1`
    - Continuum “n=2 / n=1”, :math:`N_2^{(ii)}/N_11
    - Neutral “n=3 / n=1”, :math:`N_3^{(i)}/N_1`
    - Continuum “n=3 / n=1”, :math:`N_3^{(ii)}/N_1`
    
    ... and similarly for n=4 to 9. Refer to the DEGAS2 manual for details.
    """

    def __init__(self, filepath=None):
        """Load ehr5.dat file, either from the indicated path or by downloading it locally. 

        Keyword Args:
            filepath : str, optional
                Path of ehr5.dat file to use. If left to None, the file is downloaded from the web
                and saved locally. 

        Results for each of the fields in the `fields` attribute will be available in the `res`
        attribute in the form of a dictionary. Refer to the DEGAS2 manual for a description of these
        fields.
        """

        if filepath is None:
            if not os.path.exists(atomic_data_dir + "/ehr5.dat"):
                # if ehr5.dat file is not available, download it
                download_ehr5_file()
            self.filepath = atomic_data_dir + "/ehr5.dat"
        else:
            self.filepath = filepath

        self.ne = 10 ** np.array(
            [10 + (jn - 1.0) / 2.0 for jn in np.arange(1, 16)]
        )  # cm^{-3}
        self.Te = 10 ** np.array(
            [-1.2 + (jt - 1.0) / 10.0 for jt in np.arange(1, 61)]
        )  # eV

        self.fields = [
            "Seff",
            "Reff",
            "Ei_loss",
            "Eii_loss",
            "n3i_n1",
            "n3ii_n1",
            "n2i_n1",
            "n2ii_n1",
            "n4i_n1",
            "n4ii_n1",
            "n5i_n1",
            "n5ii_n1",
            "n6i_n1",
            "n6ii_n1",
            "n7i_n1",
            "n7ii_n1",
            "n8i_n1",
            "n8ii_n1",
            "n9i_n1",
            "n9ii_n1",
        ]

        # get data
        self.load()

    def load(self):

        self.res = {}
        with open(self.filepath) as f:
            for field in self.fields:

                data = np.zeros((15, 60))

                # read header
                header = f.readline()

                for jn in np.arange(15):
                    # loop over 15 densities
                    _jn_index = f.readline()
                    arr = []
                    for jt_row in np.arange(10):
                        # 10 rows of 6 values each for Te
                        elems = [
                            val for val in f.readline().strip().split(" ") if val != ""
                        ]

                        line = [float(val) for val in elems]
                        data[jn, jt_row * 6 : (jt_row + 1) * 6] = np.array(line)
                    _dum = f.readline()  # empty line at the end

                self.res[field] = copy.deepcopy(data)

    def plot(self, field="Seff", fig=None, axes=None):
        colormap = cm.rainbow
        if fig is None or axes is None:
            fig, ax = plt.subplots()

        ls_cycle = plot_tools.get_ls_cycle()

        labels = ["{:.2e}".format(val) + " cm${-3}$" for val in self.ne]
        for i in np.arange(len(labels)):
            ax.plot(
                np.log10(self.Te),
                self.res[field][i, :],
                next(ls_cycle),
                label=labels[i],
            )

        ax.set_xlabel("$\log T_e\ \mathrm{[eV]}$", fontsize=16)
        ax.set_ylabel(field, fontsize=16)
        ax.legend()


def get_exc_state_ratio(
    m, N1, ni, ne, Te, rad_prof=None, rad_label=r"rmin [cm]", plot=False
):
    r"""Compute density of excited states in state `m` (m>1), given the density of ground state atoms.
    This function is not l-resolved.

    The function returns

    .. math::

        N_m/N_1 = \left(\frac{N_m^i}{N_1} \right) N_m + \left(\frac{N_m^{ii}}{n_i} \right) n_i


    where :math:`N_m` is the number of electrons in the excited state :math:`m`, :math:`N_1` 
    is the number in the ground state, and :math:`n_i` is the density of ions that could recombine.
    :math:`i` and :math:`ii` indicate terms corresponding to coupling to the ground state and to 
    the continuum, respectively.

    Ref.: DEGAS2 manual.

    Parameters
    ----------
    m : int
        Principal quantum number of excited state of interest. 2<m<10
    N1 : float, list or 1D-array [:math:`cm^{-3}`]
        Density of ions in the ground state. This must have the same shape as ni!
    ni : float, list or 1D-array [:math:`cm^{-3}`]
        Density of ions corresponding to the atom under consideration. This must
        have the same shape as N1!
    ne : float, list or 1D-array [:math:`cm^{-3}`]
        Electron density to evaluate atomic rates at.
    Te : float, list or 1D-array [:math:`eV`]
        Electron temperature to evaluate atomic rates at.
    rad_prof : list, 1D array or None
        If None, excited state densities are evaluated at all the combinations of ne,Te and zip(Ni,ni).
        If a 1D array (same length as ne,Te,ni and N1), then this is taken to be a radial coordinate
        for radial profiles of ne,Te,ni and N1.
    rad_label : str
        When rad_prof is not None, this is the label for the radial coordinate.
    plot : bool
        Display the excited state ratio

    Returns
    -------
    Nm : array of shape [len(ni)=len(N1),len(ne),len(Te)]
        Density of electrons in excited state `n`  [:math:`cm^{-3}`]
    """
    if m <= 1:
        raise ValueError(
            "Excited state principal quantum number must be greater than 1!"
        )
    if m > 9:
        raise ValueError("Selected excited state value not available!")

    ne = np.atleast_1d(ne)
    Te = np.atleast_1d(Te)
    ni = np.atleast_1d(ni)
    N1 = np.atleast_1d(N1)

    if rad_prof is not None:
        # if a radial profile is being computed, ne, Te, ni and N1 must all have the same length
        assert len(ne) == len(Te) and len(ne) == len(ni) and len(ne) == len(N1)

    # get CR model results:
    atom = ehr5_file()

    # coupling to the ground state:
    ground_coupling = atom.res["n{}i_n1".format(m)]
    cont_coupling = atom.res["n{}ii_n1".format(m)]

    if rad_prof is not None:
        # evaluate along radial profiles
        gc_interp = interp2d(atom.ne, atom.Te, ground_coupling.T)
        gc = np.array([float(gc_interp(XX, YY)) for XX, YY in zip(ne, Te)])
        cc_interp = interp2d(atom.ne, atom.Te, cont_coupling.T)
        cc = np.array([float(cc_interp(XX, YY)) for XX, YY in zip(ne, Te)])
    else:
        # evaluate at all combinations of points
        gc = interp2d(atom.ne, atom.Te, ground_coupling.T)(ne, Te).T
        cc = interp2d(atom.ne, atom.Te, cont_coupling.T)(ne, Te).T

        N1 = np.rollaxis(np.tile(N1, (cc.shape[0], cc.shape[1], 1)), axis=2)
        ni = np.rollaxis(np.tile(ni, (cc.shape[0], cc.shape[1], 1)), axis=2)
        gc = np.tile(gc, (len(N1), 1, 1))
        cc = np.tile(cc, (len(N1), 1, 1))

    # combine coupling to ground state and to continuum
    Nm = gc * N1 + cc * ni
    Nm_ground = gc * N1
    Nm_cont = cc * ni

    if plot:
        # plot only the first provided value of value of N1 and ni
        ls_style = plot_tools.get_ls_cycle()

        if rad_prof is not None:
            fig, ax = plt.subplots()
            ax.plot(rad_prof, Nm / N1, next(ls_style), lw=2)
            ax.set_ylabel(r"$N_{}/N_1$".format(m), fontsize=16)
            ax.set_xlabel(rad_label, fontsize=16)
        else:
            fig, ax = plt.subplots(1, 2, figsize=(15, 8))

            labels_Te = ["{:.2e}".format(val) + " eV" for val in Te]
            for jt in np.arange(len(Te)):
                ax[0].semilogx(
                    ne, Nm[0, :, jt] / N1[0], next(ls_style), lw=2, label=labels_Te[jt]
                )
            ax[0].legend(loc="best")
            ax[0].set_title("$N_1$=%.2e m$^{-3}$, $n_i$=%.2e m$^{-3}$" % (N1[0], ni[0]))
            ax[0].set_ylabel(r"$N_{}/N_1$".format(m), fontsize=16)
            ax[0].set_xlabel(r"$n_e$ [cm$^{-3}$]", fontsize=16)

            labels_ne = ["{:.2e}".format(val) + " cm$^{-3}$" for val in ne]
            for jn in np.arange(len(ne)):
                ax[1].semilogx(
                    Te, Nm[0, jn, :] / N1[0], next(ls_style), lw=2, label=labels_ne[jn]
                )
            ax[1].legend(loc="best")
            ax[1].set_title("$N_1$=%.2e m$^{-3}$, $n_i$=%.2e m$^{-3}$" % (N1[0], ni[0]))

            ax[1].set_xlabel(r"$T_e$ [eV]", fontsize=16)

        plt.tight_layout()

    return Nm, Nm_ground, Nm_cont


def plot_exc_ratios(
    n_list=[2, 3, 4, 5, 6, 7, 8, 9],
    ne=1e13,
    ni=1e13,
    Te=50,
    N1=1e12,
    ax=None,
    ls="-",
    c="r",
    label=None,
):
    """Plot :math:`N_i/N_1`, the ratio of hydrogen neutral density in the excited state `i`
    and the ground state, for given electron density and temperature.

    Parameters
    ----------
    n_list : list of integers
        List of excited states (principal quantum numbers) to consider.
    ne : float
        Electron density in :math:`cm^{-3}`.
    ni : float
        Ionized hydrogen density [:math:`cm^{-3}`]. This may be set equal to ne for a pure plasma.
    Te : float
        Electron temperature in :math:`eV`.
    N1 : float
        Density of ground state hydrogen [:math:`cm^{-3}`]. This is needed because the excited
        state fractions depend on the balance of excitation from the ground state and
        coupling to the continuum.
    ax : matplotlib.axes instance, optional
        Axes instance on which results should be plotted. 
    ls : str
        Line style to use
    c : str or other matplotlib color specification
        Color to use in plots
    label : str
        Label to use in scatter plot. 

    Returns
    -------
    Ns : list of arrays
        List of arrays for each of the n-levels requested, each containing excited state 
        densities at the chosen densities and temperatures for the given ground state density.
    """

    Ns = np.zeros(len(n_list))
    for i, n in enumerate(n_list):
        Ns[i], _, _ = get_exc_state_ratio(m=n, N1=N1, ni=ne, ne=ne, Te=Te, plot=False)

    if ax is None:
        fig, ax = plt.subplots()

    ax.scatter(n_list, Ns / N1, c=c, label=label, s=50.0)
    ax.set_xlabel("n", fontsize=16)
    ax.set_ylabel(r"$N_i/N_1$", fontsize=16)
    ax.set_ylim([0, np.max(Ns / N1) * 1.1])

    return Ns


def Lya_to_neut_dens(
    emiss_prof, ne, Te, ni=None, plot=True, rhop=None, rates_source="adas", axs=None
):
    """ Estimate ground state neutral density from measured emissivity profiles.
    This ignores possible molecular dynamics and effects that may be captured via
    forward modeling of neutral transport.

    Parameters
    ----------
    emiss_prof : 1D array
        Emissivity profile, units of :math:`W/cm`
    ne : 1D array
        Electron density, units of :math:`cm^{-3}`
    Te : 1D array
        Electron temperature, units of :math:`eV`
    ni : 1D array
        Main ion (H/D/T) density, units of :math:`cm^{-3}`. 
        If left to None, this is internally set to ni=ne.
    plot : bool
        If True, plot some of the key density profiles.
    rhop : 1D array
        Sqrt of normalized poloidal flux radial coordinate. Used only for plotting.
    rates_source : str
        Source of atomic rates. Possible choices are 'adas' or 'colrad'
    axs : Axes instance
        If given, plot on these axes.

    Returns
    ------------
    N1 : 1D array
        Radial profile of estimated ground state atomic neutral density on the same grid 
        as the input arrays. Units of :math:`cm^{-3}`.

    Examples
    --------------
    
    >>> N2_colrad,axs = Lya_to_neut_dens_basic(emiss_prof, ne, Te, ni, 
    >>>                                  plot=True, rhop=rhop, rates_source='colrad')
    
    >>> N2_adas,axs = Lya_to_neut_dens_basic(emiss_prof, ne, Te, ni, plot=True, rhop=rhop, 
    >>>                       rates_source='adas',axs=axs)
    
    """
    assert len(emiss_prof) == len(ne) and len(ne) == len(Te)
    if ni is None:
        ni = copy.deepcopy(ne)
    else:
        assert len(ne) == len(ni)

    thirteenpointsix = h * c_light * Rydberg / e
    A_21 = 4.699e8  # s^-1
    E_21 = thirteenpointsix * (1.0 - 2.0 ** (-2.0)) * e  # J

    # n=2 population in cm^-3
    N2 = emiss_prof / (A_21 * E_21)

    if rates_source == "colrad":
        # use rates from COLRAD code
        atom = ehr5_file()
        ground_coupling = atom.res["n2i_n1"]
        cont_coupling = atom.res["n2ii_n1"]

        # atomic data is on a grid in units of cm^-3, eV
        gc_interp = interp2d(atom.ne, atom.Te, ground_coupling.T)
        gc = np.array(
            [float(gc_interp(XX, YY)) for XX, YY in zip(ne, Te)]
        )  # Nm/N1 from exc
        cc_interp = interp2d(atom.ne, atom.Te, cont_coupling.T)
        cc = np.array(
            [float(cc_interp(XX, YY)) for XX, YY in zip(ne, Te)]
        )  # Nm/N1 from recomb

        # ground state density
        # N1 = (N2 - cc * ni)/gc
        N1 = (
            N2 / gc
        )  # continuum recomb term not well constrained, but should be small. Ignore it

    elif rates_source == "adas":
        filename = "pec96#h_pju#h0.dat"  # for D Ly-alpha

        # fetch file automatically, locally, from AURORA_ADAS_DIR, or directly from the web:
        path = adas_files.get_adas_file_loc(filename, filetype="adf15")
        log10pec_dict = radiation.read_adf15(path)  # , plot_lines=[1215.2])

        # evaluate these interpolations on our profiles
        pec_recomb = 10 ** log10pec_dict[1215.2]["recom"].ev(np.log10(ne), np.log10(Te))
        pec_exc = 10 ** log10pec_dict[1215.2]["excit"].ev(np.log10(ne), np.log10(Te))

        N1 = emiss_prof / E_21 / (ne * pec_exc + ni * pec_recomb)

    if plot:
        if rhop is None:
            print("No rhop array was given!")
            rhop = np.arange(len(N1))
        if axs is None:
            fig, ax = plt.subplots(2, 1, figsize=(10, 7))
        else:
            ax = axs
        sel = np.argmin(np.abs(rhop - 0.9))
        if axs is None:
            # plot this only the first time
            ax[0].plot(rhop[sel:], N2[sel:], label=r"$N_2$")
        ax[0].semilogy(rhop[sel:], N1[sel:], label=fr"{rates_source} $N_1$")
        ax[1].semilogy(
            rhop[sel:], N1[sel:] / ne[sel:], label=fr"{rates_source} $N_1/n_e$"
        )
        ax[0].set_ylim([1e10, None])
        ax[1].set_ylim([1e-5, None])
        ax[0].set_ylabel(r"Neutral density [$cm^{{-3}}$]")
        ax[1].set_ylabel(r"Density ratio")
        ax[0].legend()
        ax[1].legend()
        ax[1].set_xlabel(r"$\rho_p$")
    else:
        ax = None

    return N1, ax



"""
Script to demonstrate use of superstages.

It is recommended to run this in IPython.
"""

import numpy as np
import matplotlib.pyplot as plt

plt.ion()
from omfit_classes import omfit_eqdsk
import sys, os
from scipy.interpolate import interp1d

# Make sure that package home is added to sys.path
sys.path.append("../")
import aurora

try:  # pass any argument via the command line to show plots
    plot = len(sys.argv) > 1
except:
    plot = False

# read in default Aurora namelist
namelist = aurora.default_nml.load_default_namelist()

# Use gfile and statefile in local directory:
examples_dir = os.path.dirname(os.path.abspath(__file__))
geqdsk = omfit_eqdsk.OMFITgeqdsk(examples_dir + "/example.gfile")


# save kinetic profiles on a rhop (sqrt of norm. pol. flux) grid
# parameterization f=(f_center-f_edge)*(1-rhop**alpha1)**alpha2 + f_edge
kp = namelist["kin_profs"]
T_core = 5e3  # eV
T_edge = 100  # eV
T_alpha1 = 2.0
T_alpha2 = 1.5
n_core = 1e14  # cm^-3
n_edge = 0.4e14  # cm^-3
n_alpha1 = 2
n_alpha2 = 0.5

rhop = kp["Te"]["rhop"] = kp["ne"]["rhop"] = np.linspace(0, 1, 100)
kp["ne"]["vals"] = (n_core - n_edge) * (1 - rhop ** n_alpha1) ** n_alpha2 + n_edge
kp["Te"]["vals"] = (T_core - T_edge) * (1 - rhop ** T_alpha1) ** T_alpha2 + T_edge

# set impurity species and sources rate
imp = namelist["imp"] = "Ar"
namelist["source_type"] = "const"
namelist["source_rate"] = 2e20  # particles/s

# selection of superstages for Ar
superstages = [0, 14, 15, 16, 17, 18]

########
# first run WITHOUT superstages
namelist["superstages"] = []

# Now get aurora setup
asim = aurora.core.aurora_sim(namelist, geqdsk=geqdsk)

# set time-independent transport coefficients (flat D=1 m^2/s, V=0 cm/s)

D_z = 1e4 * np.ones(len(asim.rvol_grid))  # cm^2/s
V_z = 0.0 * np.ones(len(asim.rvol_grid))  # cm/s

# run Aurora forward model and plot results
out = asim.run_aurora(D_z, V_z, times_DV=[1.0,], unstage=True, plot=plot)

# extract densities and particle numbers in each simulation reservoir
nz, N_wall, N_div, N_pump, N_ret, N_tsu, N_dsu, N_dsul, rcld_rate, rclw_rate = out


# plot charge state distributions over radius and time
aurora.plot_tools.slider_plot(
    asim.rhop_grid,
    asim.time_grid,
    nz.transpose(1, 0, 2),
    xlabel=r"$\rho_p$",
    ylabel="time [s]",
    zlabel=r"$n_z$ [$cm^{-3}$]",
    labels=[str(i) for i in np.arange(0, nz.shape[1])],
    plot_sum=True,
)  # , x_line=asim.rvol_lcfs)


########
# now choose superstages: always include 0 and 1!
namelist["superstages"] = superstages

# set up aurora again, this time with superstages
asim = aurora.core.aurora_sim(namelist, geqdsk=geqdsk)

# run Aurora forward model and plot results
out = asim.run_aurora(D_z, V_z, times_DV=[1.0,], unstage=True, plot=plot)

# extract densities and particle numbers in each simulation reservoir
nzs, N_wall, N_div, N_pump, N_ret, N_tsu, N_dsu, N_dsul, rcld_rate, rclw_rate = out

# plot charge state distributions over radius and time
aurora.plot_tools.slider_plot(
    asim.rvol_grid,
    asim.time_grid,
    nzs.transpose(1, 0, 2),
    xlabel=r"$\rho_p$",
    ylabel="time [s]",
    zlabel=r"$n_z$ [$cm^{-3}$]",
    labels=[str(i) for i in np.arange(0, nzs.shape[1])],
    plot_sum=True,
)  # , x_line=asim.rvol_lcfs)


# compare at last slice
ls_cycle = aurora.get_ls_cycle()

fig = plt.figure()
fig.set_size_inches(12, 7, forward=True)
a_plot = plt.subplot2grid((10, 10), (0, 0), rowspan=10, colspan=8, fig=fig)
a_legend = plt.subplot2grid((10, 10), (0, 8), rowspan=10, colspan=2, fig=fig)
a_legend.axis("off")

for cs in np.arange(nz.shape[1]):
    ls = next(ls_cycle)
    a_plot.plot(asim.rhop_grid, nz[:, cs, -1], ls, lw=1.0)
    a_plot.plot(asim.rhop_grid, nzs[:, cs, -1], ls, lw=2.0)
    a_legend.plot([], [], ls, label=imp + f"$^{{{cs}+}}$")

a_plot.set_xlabel(r"$\rho_p$")
a_plot.set_ylabel(r"$n_z$ [A.U.]")

a_legend.legend(loc="best", ncol=1).set_draggable(True)



"""
Script to test functionality from namelist creation to run and postprocessing.

It is recommended to run this in IPython.
"""
import numpy as np
import matplotlib.pyplot as plt
from omfit_classes import omfit_eqdsk
import sys
import os

# Make sure that package home is added to sys.path
sys.path.append("../")
import aurora

plt.ion()

# pass any argument via the command line to show plots
plot = len(sys.argv) > 1

# read in default Aurora namelist
namelist = aurora.default_nml.load_default_namelist()

# Use gfile and statefile in local directory:
examples_dir = os.path.dirname(os.path.abspath(__file__))
geqdsk = omfit_eqdsk.OMFITgeqdsk(examples_dir + "/example.gfile")

# save kinetic profiles on a rhop (sqrt of norm. pol. flux) grid
# parameterization f=(f_center-f_edge)*(1-rhop**alpha1)**alpha2 + f_edge
kp = namelist["kin_profs"]
T_core = 5e3  # eV
T_edge = 100  # eV
T_alpha1 = 2.0
T_alpha2 = 1.5
n_core = 1e14  # cm^-3
n_edge = 0.4e14  # cm^-3
n_alpha1 = 2
n_alpha2 = 0.5

rhop = kp["Te"]["rhop"] = kp["ne"]["rhop"] = np.linspace(0, 1, 100)
kp["ne"]["vals"] = (n_core - n_edge) * (1 - rhop ** n_alpha1) ** n_alpha2 + n_edge
kp["Te"]["vals"] = (T_core - T_edge) * (1 - rhop ** T_alpha1) ** T_alpha2 + T_edge

# set impurity species and sources rate
imp = namelist["imp"] = "Ar"
namelist["source_type"] = "const"
namelist["source_rate"] = 2e20  # particles/s

# Now get aurora setup
asim = aurora.core.aurora_sim(namelist, geqdsk=geqdsk)

# set time-independent transport coefficients (flat D=1 m^2/s, V=-2 cm/s)
D_z = 1e4 * np.ones(len(asim.rvol_grid))  # cm^2/s
V_z = -2e2 * np.ones(len(asim.rvol_grid))  # cm/s

# run Aurora forward model and plot results
out = asim.run_aurora(D_z, V_z, plot=plot)

# extract densities and particle numbers in each simulation reservoir
nz, N_wall, N_div, N_pump, N_ret, N_tsu, N_dsu, N_dsul, rcld_rate, rclw_rate = out

if plot:
    # plot charge state distributions over radius and time
    aurora.plot_tools.slider_plot(
        asim.rvol_grid,
        asim.time_out,
        nz.transpose(1, 0, 2),
        xlabel=r"$r_V$ [cm]",
        ylabel="time [s]",
        zlabel=r"$n_z$ [$cm^{-3}$]",
        labels=[str(i) for i in np.arange(0, nz.shape[1])],
        plot_sum=True,
        x_line=asim.rvol_lcfs,
    )

# add radiation
asim.rad = aurora.compute_rad(
    imp,
    nz.transpose(2, 1, 0),
    asim.ne,
    asim.Te,
    prad_flag=True,
    thermal_cx_rad_flag=False,
    spectral_brem_flag=False,
    sxr_flag=False,
)

if plot:
    # plot radiation profiles over radius and time
    aurora.slider_plot(
        asim.rvol_grid,
        asim.time_out,
        asim.rad["line_rad"].transpose(1, 2, 0),
        xlabel=r"$r_V$ [cm]",
        ylabel="time [s]",
        zlabel=r"Line radiation [$MW/m^3$]",
        labels=[str(i) for i in np.arange(0, nz.shape[1])],
        plot_sum=True,
        x_line=asim.rvol_lcfs,
    )



# MIT License
#
# Copyright (c) 2021 Francesco Sciortino
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
from matplotlib import pyplot as plt
from matplotlib import animation

plt.ion()
import pickle as pkl
from . import plot_tools


def animate_aurora(
    x,
    y,
    z,
    xlabel="",
    ylabel="",
    zlabel="",
    labels=None,
    plot_sum=False,
    uniform_y_spacing=True,
    save_filename=None,
):
    """ Produce animation of time- and radially-dependent results from aurora.

    Parameters
    ----------
    x : array of float, (`M`,)
        The abscissa. (in aurora, often this may be rhop)
    y : array of float, (`N`,)
        The variable to slide over. (in aurora, often this may be time)
    z : array of float, (`P`, `M`, `N`)
        The variables to plot.
    xlabel : str, optional
        The label for the abscissa.
    ylabel : str, optional
        The label for the animated coordinate. This is expected in a format such that ylabel.format(y_val)
        will display a good moving label, e.g. ylabel='t={:.4f} s'.
    zlabel : str, optional
        The label for the ordinate.
    labels : list of str with length `P`
        The labels for each curve in `z`.
    plot_sum : bool, optional
        If True, will also plot the sum over all `P` cases. Default is False.
    uniform_y_spacing : bool, optional
        If True, interpolate values in z onto a uniformly-spaced y grid
    save_filename : str
        If a valid path/filename is provided, the animation will be saved here in mp4 format. 
    """
    if labels is None:
        labels = ["" for v in z]

    if plot_sum:
        labels.append("total")
        z_sum = np.sum(z, axis=0)
        z = np.vstack((z, np.atleast_3d(z_sum).transpose(2, 0, 1)))

    if uniform_y_spacing:
        from scipy.interpolate import RegularGridInterpolator as rgi

        interp_fun = rgi((np.arange(z.shape[0]), x, y), z)

        y_eq = np.linspace(min(y), max(y), len(y))
        new_grid = np.ix_(np.arange(z.shape[0]), x, y_eq)
        z = interp_fun(new_grid)

    # set up the figure and side space for legend
    fig = plt.figure(figsize=(10, 6))
    a_plot = plt.subplot2grid((1, 10), (0, 0), rowspan=1, colspan=8, fig=fig)
    a_legend = plt.subplot2grid((1, 10), (0, 8), rowspan=1, colspan=2, fig=fig)
    a_plot.set_xlabel(xlabel)
    a_plot.set_ylabel(zlabel)

    a_plot.set_xlim([np.min(x), np.max(x)])
    a_plot.set_ylim([0.0, 1.1 * np.max(z)])

    # get nice sequence of line styles/colors
    ls_cycle = plot_tools.get_ls_cycle()

    lines = []
    for l_ in labels:
        a_legend.plot(
            [],
            [],
            "k-" if l_ == "total" else next(ls_cycle),
            lw=2.5 if l_ == "total" else 1.0,
            label=l_,
        )[0]
        lobj = a_plot.plot(
            [],
            [],
            "k-" if l_ == "total" else next(ls_cycle),
            lw=2.5 if l_ == "total" else 1.0,
        )[0]
        lines.append(lobj)

    # time label (NB: update won't work if this is placed outside axes)
    y_text = a_plot.text(0.75, 0.95, " ", fontsize=14, transform=a_plot.transAxes)

    def init():  # initialization function
        for line in lines:
            line.set_data([], [])
        y_text.set_text("")
        return tuple(lines) + (y_text,)

    def animate(i):  # animation function, called sequentially
        y_text.set_text(ylabel.format(y_eq[i] if uniform_y_spacing else y[i]))

        for lnum, line in enumerate(lines):
            line.set_data(x, z[lnum, :, i])
        return tuple(lines) + (y_text,)

    a_legend.legend(loc="center").set_draggable(True)
    a_legend.axis("off")

    # run animation now:
    anim = animation.FuncAnimation(
        fig, animate, init_func=init, frames=len(y), interval=20, blit=True
    )

    if save_filename is not None:
        if "gif" in save_filename:
            anim.save(save_filename, fps=30)
        else:
            anim.save(save_filename + ".mp4", fps=30, extra_args=["-vcodec", "libx264"])



"""
Script to test functionality for AUG.

It is recommended to run this in IPython.
"""

import numpy as np
import matplotlib.pyplot as plt

plt.ion()
import sys, os
from scipy.interpolate import interp1d
import aug_sfutils as sf
from omfit_classes.omfit_eqdsk import OMFITgeqdsk

# Make sure that package home is added to sys.path
sys.path.append("../")
import aurora

shot = 39649
time = 3.0

# read in default Aurora namelist
namelist = aurora.default_nml.load_default_namelist()

# get equilibrium for AUG through aug_sfutils and OMFITgeqdsk
geqdsk = OMFITgeqdsk("").from_aug_sfutils(shot=shot, time=time, eq_shotfile="EQI")

# get ne, Te from AUG IDA at specified time
kp = namelist["kin_profs"]

ida = sf.SFREAD(shot, "ida")
time_ida = ida.gettimebase("Te")
it_ida = np.argmin(np.abs(time_ida - time))
rhop_ida = ida.getareabase("Te")
Te_eV = ida.getobject("Te")
ne_m3 = ida.getobject("ne")

kp["Te"]["rhop"] = kp["ne"]["rhop"] = rhop_ida[:, it_ida]
kp["ne"]["vals"] = ne_m3[:, it_ida] * 1e-6  # m^-3 --> cm^-3
kp["Te"]["vals"] = Te_eV[:, it_ida]  # eV

# set impurity species and sources rate
imp = namelist["imp"] = "Ar"
namelist["source_type"] = "const"
namelist["source_rate"] = 2e20  # particles/s

# Now get aurora setup
asim = aurora.core.aurora_sim(namelist, geqdsk=geqdsk)

# set time-independent transport coefficients (flat D=1 m^2/s, V=-2 cm/s)
D_z = 1e4 * np.ones(len(asim.rvol_grid))  # cm^2/s
V_z = -2e2 * np.ones(len(asim.rvol_grid))  # cm/s

# run Aurora forward model and plot results
out = asim.run_aurora(D_z, V_z, plot=True)

# extract densities and particle numbers in each simulation reservoir
nz, N_wall, N_div, N_pump, N_ret, N_tsu, N_dsu, N_dsul, rcld_rate, rclw_rate = out

# plot charge state distributions over radius and time
aurora.plot_tools.slider_plot(
    asim.rvol_grid,
    asim.time_out,
    nz.transpose(1, 0, 2),
    xlabel=r"$r_V$ [cm]",
    ylabel="time [s]",
    zlabel=r"$n_z$ [$cm^{-3}$]",
    labels=[str(i) for i in np.arange(0, nz.shape[1])],
    plot_sum=True,
    x_line=asim.rvol_lcfs,
)

# add radiation
asim.rad = aurora.compute_rad(
    imp,
    nz.transpose(2, 1, 0),
    asim.ne,
    asim.Te,
    prad_flag=True,
    thermal_cx_rad_flag=False,
    spectral_brem_flag=False,
    sxr_flag=False,
)

# plot radiation profiles over radius and time
aurora.slider_plot(
    asim.rvol_grid,
    asim.time_out,
    asim.rad["line_rad"].transpose(1, 2, 0),
    xlabel=r"$r_V$ [cm]",
    ylabel="time [s]",
    zlabel=r"Line radiation [$MW/m^3$]",
    labels=[str(i) for i in np.arange(0, nz.shape[1])],
    plot_sum=True,
    x_line=asim.rvol_lcfs,
)



"""Cooling curves for multiple ions from Aurora. 

Here we exclude charge exchange for simplicity, but it may easily be added by specifying background D neutral densities.
Note that the cooling factors shown here only use ionization equilibrium, i.e. no transport effects.

sciortino, 2021
"""

import aurora
import numpy as np
import matplotlib.pyplot as plt

plt.ion()


# scan Te and fix a value of ne
Te_eV = np.logspace(np.log10(100), np.log10(1e5), 1000)
ne_cm3 = 5e13 * np.ones_like(Te_eV)

imp = "C"

# basic cooling curve, considering ionization equilibrium between charge states
line_rad_tot, cont_rad_tot = aurora.get_cooling_factors(
    imp, ne_cm3, Te_eV, plot=True, ion_resolved=False
)


# Total cooling coefficients are the same regardless of whether we apply superstaging or not!
line_rad_tot, cont_rad_tot = aurora.get_cooling_factors(
    imp, ne_cm3, Te_eV, plot=True, ion_resolved=False, superstages=[0, 3, 6]
)


# plot contributions from each charge state at ionization equilibrium
line_rad_tot, cont_rad_tot = aurora.get_cooling_factors(
    imp, ne_cm3, Te_eV, plot=True, ion_resolved=True
)


###### Overplot total cooling coefficients for multiple ions
ions_list = ["He", "C", "Ar", "W"]

fig = plt.figure()
a_plot = plt.subplot2grid((10, 10), (0, 0), rowspan=10, colspan=8, fig=fig)
a_legend = plt.subplot2grid((10, 10), (0, 8), rowspan=10, colspan=8, fig=fig)

a_legend.axis("off")

ls_cycle = aurora.get_ls_cycle()
for imp in ions_list:

    ls = next(ls_cycle)

    # read atomic data, interpolate and plot cooling factors
    line_rad_tot, cont_rad_tot = aurora.get_cooling_factors(
        imp, ne_cm3, Te_eV, ion_resolved=False, plot=False
    )

    # total radiation (includes hard X-ray, visible, UV, etc.)
    a_plot.loglog(Te_eV / 1e3, cont_rad_tot + line_rad_tot, ls)
    a_legend.plot([], [], ls, label=f"{imp}")


a_legend.legend(loc="best").set_draggable(True)
a_plot.grid("on", which="both")
a_plot.set_xlabel("T$_e$ [keV]")
a_plot.set_ylabel("$L_z$ [$W$ $m^3$]")
plt.tight_layout()



"""
Script to test functionality of steady-state run with AURORA.

It is recommended to run this in IPython.
"""

import numpy as np
import matplotlib.pyplot as plt

plt.ion()
from omfit_classes import omfit_eqdsk
import sys, os
from scipy.interpolate import interp1d
import copy

# Make sure that package home is added to sys.path
sys.path.append("../")
import aurora

# read in default Aurora namelist
namelist = aurora.default_nml.load_default_namelist()

# Use gfile and statefile in local directory:
examples_dir = os.path.dirname(os.path.abspath(__file__))
geqdsk = omfit_eqdsk.OMFITgeqdsk(examples_dir + "/example.gfile")

# save kinetic profiles on a rhop (sqrt of norm. pol. flux) grid
# parameterization f=(f_center-f_edge)*(1-rhop**alpha1)**alpha2 + f_edge
kp = namelist["kin_profs"]
T_core = 5e3  # eV
T_edge = 100  # eV
T_alpha1 = 2.0
T_alpha2 = 1.5
n_core = 1e14  # cm^-3
n_edge = 0.4e14  # cm^-3
n_alpha1 = 2
n_alpha2 = 0.5

rhop = kp["Te"]["rhop"] = kp["ne"]["rhop"] = np.linspace(0, 1, 100)
ne = kp["ne"]["vals"] = (n_core - n_edge) * (1 - rhop ** n_alpha1) ** n_alpha2 + n_edge
Te = kp["Te"]["vals"] = (T_core - T_edge) * (1 - rhop ** T_alpha1) ** T_alpha2 + T_edge

# set impurity species and sources rate to 0
imp = namelist["imp"] = "Ar"
namelist["source_type"] = "const"
namelist["source_rate"] = 1e31  # particles/s

# get charge state distributions from ionization equilibrium
atom_data = aurora.atomic.get_atom_data(imp, ["scd", "acd"])

# get fractional abundances on ne (cm^-3) and Te (eV) grid
_Te, fz = aurora.atomic.get_frac_abundances(atom_data, ne, Te, rho=rhop, plot=False)

# initial guess for steady state Ar charge state densities
nz_init = 1e-25 * ne[:, None] * fz

# Now get aurora setup
asim = aurora.core.aurora_sim(namelist, geqdsk=geqdsk)

nz_init = interp1d(rhop, nz_init, bounds_error=False, fill_value=0.0, axis=0)(
    asim.rhop_grid
)

# set time-independent transport coefficients
D_z = 1e4 * np.ones(len(asim.rvol_grid))  # cm^2/s
V_z = -10e2 * asim.rhop_grid ** 5  # cm/s

n_steps = 10
max_sim_time = 500e-3

nz_norm_steady = asim.run_aurora_steady(
    D_z,
    V_z,
    nz_init=None,  # nz_init,
    tolerance=0.01,
    max_sim_time=max_sim_time,
    dt=1e-4,
    dt_increase=1.05,
    n_steps=n_steps,
    plot=True,
)

print("tau_imp:", asim.tau_imp)



"""
Script to test functionality from namelist creation to run and postprocessing.

It is recommended to run this in IPython.
"""

import numpy as np
import matplotlib.pyplot as plt
from omfit_classes import omfit_eqdsk
import pickle as pkl
import scipy
import sys
import os
import time
from copy import deepcopy

plt.ion()

# Make sure that package home is added to sys.path
sys.path.append("../")
import aurora

# number of repetitions to accurately time runs
num = 10

# read in default Aurora namelist
namelist = aurora.default_nml.load_default_namelist()
kp = namelist["kin_profs"]

# Use gfile and statefile in local directory:
examples_dir = os.path.dirname(os.path.abspath(__file__))
geqdsk = omfit_eqdsk.OMFITgeqdsk(examples_dir + "/example.gfile")

# save kinetic profiles on a rhop (sqrt of norm. pol. flux) grid
# parameterization f=(f_center-f_edge)*(1-rhop**alpha1)**alpha2 + f_edge
kp = namelist["kin_profs"]
T_core = 5e3  # eV
T_edge = 100  # eV
T_alpha1 = 2.0
T_alpha2 = 1.5
n_core = 1e14  # cm^-3
n_edge = 0.4e14  # cm^-3
n_alpha1 = 2
n_alpha2 = 0.5

rhop = kp["Te"]["rhop"] = kp["ne"]["rhop"] = np.linspace(0, 1, 100)
ne_cm3 = (n_core - n_edge) * (1 - rhop ** n_alpha1) ** n_alpha2 + n_edge
kp["ne"]["vals"] = ne_cm3[None, :]
Te_eV = (T_core - T_edge) * (1 - rhop ** T_alpha1) ** T_alpha2 + T_edge
kp["Te"]["vals"] = Te_eV[None, :]

# set impurity species and sources rate
imp = namelist["imp"] = "C"
namelist["source_type"] = "const"
namelist["source_rate"] = 1e21  # particles/s

# Change radial resolution from default:
# namelist['dr_0']=0.3
# namelist['dr_1']=0.05

# Change time resolution from default:
# namelist['timing']['dt_increase'] = np.array([1.01, 1.])
# namelist['timing']['dt_start'] = np.array([1e-5, 0.001])
# namelist['timing']['steps_per_cycle'] = np.array([1,1])
# namelist['timing']['times'] = np.array([0.,0.2])

# Now get aurora setup (let setup process create inputs needed to produce radial and temporal grids)
asim = aurora.core.aurora_sim(namelist, geqdsk=geqdsk)

# Visualize radial grid
rvol_grid, pro_grid, qpr_grid, prox_param = aurora.grids_utils.create_radial_grid(
    asim.namelist, plot=True
)

# Visualize time resolution
time_grid, save_grid = aurora.grids_utils.create_time_grid(
    asim.namelist["timing"], plot=True
)

# set time-independent transport coefficients (flat D=1 m^2/s, V=-2 cm/s)
D_z = 1e4 * np.ones(len(asim.rvol_grid))  # cm^2/s
V_z = -2e2 * asim.rhop_grid ** 10  # cm/s, increasing towards the edge

# plot transport coefficients
fig, ax = plt.subplots(2, 1, sharex=True, figsize=(8, 8))
ax[0].plot(asim.rhop_grid, D_z / 1e4)
ax[1].plot(asim.rhop_grid, V_z / 1e2)
ax[1].set_xlabel(r"$\rho_p$")
ax[0].set_ylabel(r"$D$ [$m^2/s$]")
ax[1].set_ylabel(r"$v$ [$m/s$]")
plt.subplots_adjust(wspace=0, hspace=0)


####### Finite-differences method #########
start = time.time()
for n in np.arange(num):
    out = asim.run_aurora(D_z, V_z, alg_opt=0)
print("Average FD time per run: ", (time.time() - start) / num)
nz = out[
    0
]  # extract only charge state densities in the plasma from output -- (time,nZ,space)

# plot charge state distributions over radius and time
aurora.plot_tools.slider_plot(
    asim.rvol_grid,
    asim.time_out,
    nz.transpose(1, 0, 2),
    xlabel=r"$r_V$ [cm]",
    ylabel="time [s]",
    zlabel=r"$n_z$ [$cm^{-3}$]",
    labels=[str(i) for i in np.arange(0, nz.shape[1])],
    plot_sum=True,
    x_line=asim.rvol_lcfs,
)

# Check particle conservation
out_fd, axs = asim.check_conservation(plot=True)


####### Finite volumes (Linder) method #########
start = time.time()
for n in np.arange(num):
    out_2 = asim.run_aurora(D_z, V_z, alg_opt=1, evolneut=False)
print("Average FV time per run: ", (time.time() - start) / num)
nz_2 = out_2[0]

# plot charge state distributions over radius and time
aurora.plot_tools.slider_plot(
    asim.rvol_grid,
    asim.time_out,
    nz_2.transpose(1, 0, 2),
    xlabel=r"$r_V$ [cm]",
    ylabel="time [s]",
    zlabel=r"$n_z$ [$cm^{-3}$]",
    labels=[str(i) for i in np.arange(0, nz_2.shape[1])],
    plot_sum=True,
    x_line=asim.rvol_lcfs,
)

# Check particle conservation
out_fv, axs = asim.check_conservation(plot=True, axs=axs)


######### Finite volumes (Linder) evolving neutrals -- under development! #########
start = time.time()
for n in np.arange(num):
    out_3 = asim.run_aurora(D_z, V_z, alg_opt=1, evolneut=True)
print("Average FVN time per run: ", (time.time() - start) / num)
nz_3 = out_3[0]

# plot charge state distributions over radius and time
aurora.plot_tools.slider_plot(
    asim.rvol_grid,
    asim.time_out,
    nz_3.transpose(1, 0, 2),
    xlabel=r"$r_V$ [cm]",
    ylabel="time [s]",
    zlabel=r"$n_z$ [$cm^{-3}$]",
    labels=[str(i) for i in np.arange(0, nz_3.shape[1])],
    plot_sum=True,
    x_line=asim.rvol_lcfs,
)

# Check particle conservation
out_fvn, axs = asim.check_conservation(plot=True, axs=axs)


###########################################################################
#
#           Compare all algorithmic choices at last time slice
#
############################################################################
labels = [fr"{imp}$^{{{str(i)}}}$" for i in np.arange(0, nz_3.shape[1])]
colors = plt.cm.rainbow(np.linspace(0, 1, nz.shape[1]))
fig = plt.figure()
fig.set_size_inches(10, 7, forward=True)
a_plot = plt.subplot2grid((10, 10), (0, 0), rowspan=10, colspan=8, fig=fig)
a_legend = plt.subplot2grid((10, 10), (0, 8), rowspan=10, colspan=2, fig=fig)
for ii, cc in zip(np.arange(nz.shape[1]), colors):
    a_plot.plot(asim.rhop_grid, nz[:, ii, -1].T, c=cc, ls="-")
    a_plot.plot(asim.rhop_grid, nz_2[:, ii, -1].T, c=cc, ls="--")

    #########
    factor = np.max(nz_2) / np.max(
        nz_3
    )  # factor needed to match neutral evolution to basic FV case
    a_plot.plot(asim.rhop_grid, factor * nz_3[:, ii, -1].T, c=cc, ls=":")
    ########
    a_legend.plot([], [], c=cc, label=labels[ii], ls="-")
a_legend.plot([], [], c="k", ls="-", lw=2, label="Original")
a_legend.plot([], [], c="k", ls="--", lw=2, label="Linder w/o n")
a_legend.plot([], [], c="k", ls=":", lw=2, label="Linder w/ n")

a_legend.legend(loc="best").set_draggable(True)
a_plot.set_xlabel(r"$\rho_p$")
a_plot.set_ylabel(r"$n_z$ [$cm^{-3}$]")
a_legend.axis("off")
fig.suptitle("Algorithm comparison")


######## images/plots of total density ############
fig, ax = plt.subplots()
pcm = ax.pcolormesh(
    asim.rhop_grid,
    asim.time_out,
    nz.sum(axis=1).T,
    cmap="plasma",
    vmax=nz.sum(axis=1)[0, :].max(),
    shading="auto",
)
fig.colorbar(pcm)
ax.set_xlabel(r"$\rho_p$")
ax.set_ylabel(r"$t$ [s]")
ax.set_title("Tot impurity density (Finite differences)")
ax.set_xlim([0.0, 1.0])

fig, ax = plt.subplots()
pcm = ax.pcolormesh(
    asim.rhop_grid,
    asim.time_out,
    nz_2.sum(axis=1).T,
    cmap="plasma",
    vmax=nz_2.sum(axis=1)[0, :].max(),
    shading="auto",
)
fig.colorbar(pcm, extend="max")
ax.set_xlabel(r"$\rho_p$")
ax.set_ylabel(r"$t$ [s]")
ax.set_title("Tot impurity density (Finite volumes)")
ax.set_xlim([0.0, 1.0])



"""
Setup for Aurora package. Basic call (install in editable mode)
pip install -e .

To install with a different Fortran compiler, use e.g.
python3 setup.py build --fcompiler=gnu95
or
python3 setup.py build --fcompiler=intelem 

It should be possible to pass any f2py flags via the command line, e.g. using
python3 setup.py build --fcompiler=intelem --opt="-fast"

"""

import setuptools
import os, sys, subprocess
from numpy.distutils.core import setup, Extension

package_name = "aurorafusion"


def get_version(relpath):
    """read version info from file without importing it"""
    from os.path import dirname, join

    for line in open(join(dirname(__file__), relpath)):
        if "__version__" in line:
            if '"' in line:
                return line.split('"')[1]
            elif "'" in line:
                return line.split("'")[1]


# with open('README.rst', 'r') as fh:
#    long_description = fh.read()
# long_description='See documentation at https://aurora-fusion.readthedocs.io'

####
with open("README.rst") as f:
    long_description = "\n" + "\n".join(f.read().strip().split("\n")[2:])

# Package adas_data and subdirectories together with license and readme
package_data = ["../USER_AGREEMENT.txt", "../README.rst", "*"]
for root, dir, files in os.walk("aurora"):
    if root == "aurora":
        continue
    if "__pycache__" in root:
        continue
    package_data.append(root[len("aurora") + 1 :] + os.sep + "*")

# For Fortran code:
wrapper = Extension(
    name="aurora._aurora",
    sources=[
        "aurora/main.f90",
        "aurora/grids.f90",
        "aurora/impden.f90",
        "aurora/math.f90",
    ],
)

install_requires = open("requirements.txt").read().split("\n")

setup(
    name=package_name,
    version=get_version("aurora/__init__.py"),
    description="Aurora package for particle transport, radiation and neutrals in magnetically-confined plasmas",
    long_description=long_description,
    # long_description_content_type='text/markdown',
    url="https://github.com/fsciortino/Aurora",
    author="F. Sciortino",
    author_email="sciortino@psfc.mit.edu",
    packages=["aurora"],
    keywords="particle and impurity transport, neutrals, radiation, magnetic confinement fusion",
    package_dir={"aurora": "aurora"},
    package_data={"aurora": package_data},
    data_files=[
        (
            "aurora_examples",
            [
                "examples/basic.py",
                "examples/frac_abundances.py",
                "examples/example.gfile",
                "examples/example.input.gacode",
            ],
        )
    ],
    install_requires=install_requires,
    ext_modules=[wrapper],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)



"""
Script to benchmark basic Julia version against Fortran one. 
"""

import numpy as np
import matplotlib.pyplot as plt
from omfit_classes import omfit_eqdsk
import pickle as pkl
import scipy, sys, os
import time

# Make sure that package home is added to sys.path
sys.path.append("../")
import aurora

# read in default Aurora namelist
namelist = aurora.default_nml.load_default_namelist()
kp = namelist["kin_profs"]

# Use gfile and statefile in local directory:
examples_dir = os.path.dirname(os.path.abspath(__file__))
geqdsk = omfit_eqdsk.OMFITgeqdsk(examples_dir + "/example.gfile")


# save kinetic profiles on a rhop (sqrt of norm. pol. flux) grid
# parameterization f=(f_center-f_edge)*(1-rhop**alpha1)**alpha2 + f_edge
kp = namelist["kin_profs"]
T_core = 5e3  # eV
T_edge = 100  # eV
T_alpha1 = 2.0
T_alpha2 = 1.5
n_core = 1e14  # cm^-3
n_edge = 0.4e14  # cm^-3
n_alpha1 = 2
n_alpha2 = 0.5

rhop = kp["Te"]["rhop"] = kp["ne"]["rhop"] = np.linspace(0, 1, 100)
ne_cm3 = (n_core - n_edge) * (1 - rhop ** n_alpha1) ** n_alpha2 + n_edge
kp["ne"]["vals"] = ne_cm3[None, :]
Te_eV = (T_core - T_edge) * (1 - rhop ** T_alpha1) ** T_alpha2 + T_edge
kp["Te"]["vals"] = Te_eV[None, :]

# set impurity species and sources rate
imp = namelist["imp"] = "Ar"
namelist["source_type"] = "const"
namelist["source_rate"] = 1e24

# Now get aurora setup
asim = aurora.core.aurora_sim(namelist, geqdsk=geqdsk)

# set time-independent transport coefficients (flat D=1 m^2/s, V=-2 cm/s)
D_z = 1e4 * np.ones(len(asim.rvol_grid))  # cm^2/s
V_z = -2e2 * np.ones(len(asim.rvol_grid))  # cm/s

# do several runs in series to more accurately time runs
num = 10
start = time.time()
for i in range(num):
    pyout = out = asim.run_aurora(D_z, V_z)
print("Fortran: ", (time.time() - start) / num, " seconds on average")

# First call includes precompilation, not a good timing example. Time second set of runs!
start = time.time()
juout = asim.run_aurora(D_z, V_z, use_julia=True)
print("Julia time for first call (compiling): ", time.time() - start, " second")

start = time.time()
for i in range(num):
    juout = asim.run_aurora(D_z, V_z, use_julia=True)
print("Julia: ", (time.time() - start) / num, " seconds on average")

# check that results are the same between Fortran and Julia:
all_good = True
for i in range(len(juout)):
    if not np.allclose(pyout[i], juout[i]):
        print("Result incongruency")
        all_good = False
        break
if all_good:
    print("Fortran and Julia results are equivalent!")



"""Collection of classes and functions for loading, interpolation and processing of atomic data. 
Refer also to the adas_files.py script. 
"""
# MIT License
#
# Copyright (c) 2021 Francesco Sciortino
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import RectBivariateSpline, interp1d, RegularGridInterpolator
from matplotlib import cm
import os, sys, copy
import scipy.ndimage
from scipy.linalg import svd
from scipy import constants

from . import adas_files


def get_adas_file_types():
    """Obtain a description of each ADAS file type and its meaning in the context of Aurora.

    Returns
    ------------
    dict
        Dictionary with keys given by the ADAS file types and values giving a description for them.

    Notes
    ---------
    For background on ADAS generalized collisional-radiative modeling and data formats, refer to 
    [1]_.

    References
    -----------------
    
    .. [1] Summers et al., "Ionization state, excited populations and emission of impurities 
       in dynamic finite density plasmas: I. The generalized collisional-radiative model for 
       light elements", Plasma Physics and Controlled Fusion, 48:2, 2006

    """

    return {
        "acd": "effective recombination",
        "scd": "effective ionization",
        "prb": "continuum radiation",
        "plt": "line radiation",
        "ccd": "thermal charge exchange",
        "prc": "thermal charge exchange continuum radiation",
        "pls": "line radiation in the SXR range",
        "prs": "continuum radiation in the SXR range",
        "brs": "continuum spectral bremstrahlung",
        "fis": "sensitivity in the SXR range",
        "pbs": "impurity bremsstrahlung in SXR range, also included in prs files",
    }


class adas_file:
    """Read ADAS file in ADF11 format over the given density and temperature grids. 
    Note that such grids vary between files, and the species they refer to may too.

    Refer to ADAS documentation for details on each file.

    Parameters
    ----------
    filepath : str
        Path to location where ADAS file is located.
    """

    def __init__(self, filepath):

        self.filepath = filepath
        self.filename = filepath.split("/")[-1]
        self.file_type = self.filename[:3]

        if self.file_type not in ["brs", "sxr"]:
            self.imp = self.filename.split("_")[1].split(".")[0]

        # get data
        self.load()

        # settings for plotting
        self.n_ion = self.data.shape[0]
        self.ncol = np.ceil(np.sqrt(self.n_ion))
        self.nrow = np.ceil(self.n_ion / self.ncol)

    def load(self):

        with open(self.filepath) as f:
            header = f.readline()
            n_ions, n_ne, n_T = header.split()[:3]
            details = " ".join(header.split()[3:])

            f.readline()

            n_ions, n_ne, n_T = int(n_ions), int(n_ne), int(n_T)
            logT = []
            logNe = []
            while len(logNe) < n_ne:
                line = f.readline()
                logNe = logNe + [float(n) for n in line.split()]
            while len(logT) < n_T:
                line = f.readline()
                logT = logT + [float(t) for t in line.split()]

            logT = np.array(logT)
            logNe = np.array(logNe)

            data = []
            for i_ion in range(n_ions):
                f.readline()
                plsx = []
                while len(plsx) < n_ne * n_T:
                    line = f.readline()
                    plsx = plsx + [float(L) for L in line.split()]
                plsx = np.array(plsx).reshape(n_T, n_ne)
                data.append(np.array(plsx))

        self.logNe = logNe
        self.logT = logT
        self.data = np.array(data)

    def plot(self, fig=None, axes=None):
        """Plot data from input ADAS file. If provided, the arguments allow users to overplot
        and compare data from multiple files. 

        Parameters
        ----------
        fig : matplotlib Figure object
            If provided, add specification as to which ADAS file is being plotted.
        axes : matplotlib Axes object (or equivalent)
            If provided, plot on these axes. Note that this typically needs to be a set of axes
            for each plotted charge state. Users may want to call this function once first to get
            some axes, and then pass those same axes to a second call for another file to compare with.
        """
        if fig is None or axes is None:
            fig, axes = plt.subplots(
                int(self.ncol), int(self.nrow), sharex=True, sharey=True
            )

        axes = np.atleast_2d(axes)
        colormap = cm.rainbow
        if fig is not None:
            fig.suptitle(self.filename + "  " + get_adas_file_types()[self.file_type])

        for i, ax in enumerate(axes.flatten()):
            if i >= self.n_ion:
                break
            if all(self.data[i, :, :].std(axis=1) == 0):  # independent of density
                ax.plot(self.logT, self.data[i, :, 0])
            else:
                ax.set_prop_cycle(
                    "color", colormap(np.linspace(0, 1, self.data.shape[2]))
                )
                for idens in np.arange(len(self.logNe)):
                    ax.plot(
                        self.logT,
                        self.data[i, :, idens],
                        label=f"$n_e=10^{{{self.logNe[idens]}}}$ cm$^{{{-3}}}$",
                    )
                ax.legend(loc="best").set_draggable(True)
            ax.grid(True)
            if self.file_type != "brs":
                charge = (
                    i + 1 if self.file_type in ["scd", "prs", "ccd", "prb"] else i + 2
                )
                ax.set_title(self.imp + "$^{%d\!+}$" % (charge - 1))  # check?

        for ax in axes[-1]:
            ax.set_xlabel("$\log T_e\ \mathrm{[eV]}$")
        for ax in axes[:, 0]:
            if self.file_type in ["scd", "acd", "ccd"]:
                ax.set_ylabel("$\log(" + self.file_type + ")\ \mathrm{[cm^3/s]}$")
            elif self.file_type in ["prb", "plt", "prc", "pls", "brs", "prs"]:
                ax.set_ylabel("$\log(" + self.file_type + ")\ \mathrm{[W\cdot cm^3]}$")


def read_filter_response(filepath, adas_format=True, plot=False):
    """Read a filter response function over energy. 

    This function attempts to read the data checking for the following formats (in this order):
    
    #. The ADAS format. Typically, this data is from obtained from http://xray.uu.se and produced
       via ADAS routines. 

    #. The format returned by the `Center for X-Ray Optics website  <https://henke.lbl.gov/optical_constants/filter2.html>`__ .

    Note that filter response functions are typically a combination of a filter transmissivity 
    and a detector absorption. 

    Parameters
    ----------
    filepath : str
        Path to filter file of interest.
    plot : bool
        If True, the filter response function is plotted. 
    
    """
    E_eV = []
    response = []
    try:
        # Attempt to read ADAS format
        with open(filepath) as f:

            header = f.readline()
            num = int(header.split()[0])

            # *****
            f.readline()

            while len(E_eV) < num:
                line = f.readline()
                E_eV += [float(n) for n in line.split()]
            while len(response) < num:
                line = f.readline()
                response += [float(t) for t in line.split()]

        # energy and response function are written in natural logs
        E_eV = np.concatenate(([0.0,], np.array(np.exp(E_eV))))
        response = np.concatenate(([0.0,], np.array(np.exp(response))))
    except ValueError:
        try:
            # Attempt to read CXRO format
            with open(filepath) as f:
                contents = f.readlines()

            for line in contents[2:]:
                tmp = line.strip().split()
                E_eV.append(float(tmp[0]))
                response.append(float(tmp[1]))
            E_eV = np.concatenate(([0.0,], np.array(E_eV)))
            response = np.concatenate(([0.0,], np.array(response)))
        except ValueError:
            raise ValueError("Unrecognized filter function format...")

    if plot:
        fig, ax = plt.subplots()
        ax.semilogx(E_eV, response)
        ax.set_xlabel("Photon energy [eV]")
        ax.set_ylabel("Detector response efficiency")
        plt.tight_layout()

    return E_eV, response


def get_atom_data(imp, files=["acd", "scd"]):
    """ Collect atomic data for a given impurity from all types of ADAS files available or
    for only those requested. 

    Parameters
    ----------
    imp : str
        Atomic symbol of impurity ion.
    files : list or dict
        ADAS file types to be fetched. Default is ["acd","scd"] for effective ionization 
        and recombination rates (excluding CX) using default files, listed in :py:func:`~aurora.adas_files_adas_files_dict`.
        If users prefer to use specific files, they may pass a dictionary instead, of the form

        .. code-block:: python

            {'acd': 'acd89_ar.dat', 'scd': 'scd89_ar.dat'}

        or

        .. code-block:: python

            {'acd': 'acd89_ar.dat', 'scd': None}

        if only some of the files need specifications and others (given as None) should be 
        taken from the default files.
    
    Returns
    -------
    atom_data : dict
        Dictionary containing data for each of the requested files. 
        Each entry of the dictionary gives log-10 of ne, log-10 of Te and log-10 of the data
        as attributes atom_data[key].logNe, atom_data[key].logT, atom_data[key].data
    """

    try:
        # default files dictionary
        all_files = adas_files.adas_files_dict()[
            imp
        ]  # all default files for a given species
    except KeyError:
        raise KeyError(f"ADAS data not available for ion {imp}!")

    if isinstance(files, dict):
        # user provided files choices as a dictionary
        for filename in files:
            if files[filename] is not None:
                all_files[filename] = files[filename]

    # exclude files not of interest
    keys_to_delete = [key for key in all_files if key not in files]
    for key in keys_to_delete:
        del all_files[key]

    for filecheck in files:
        if filecheck not in all_files:
            raise ValueError(f"Could not fetch {imp} {filecheck.upper()} file!")

    atom_data = {}
    for filetype in all_files:

        # find location of required ADF11 file
        fileloc = adas_files.get_adas_file_loc(all_files[filetype], filetype="adf11")

        # load specific file and add it to output dictionary
        res = adas_file(fileloc)
        atom_data[filetype] = res.logNe, res.logT, res.data

    return atom_data


def null_space(A):
    """Find null space of matrix `A`.
    """
    u, s, vh = svd(A, full_matrices=True)
    Q = vh[-1, :].T.conj()  # -1 index is after infinite time/equilibration
    # return the smallest singular eigenvalue
    return Q, s[-2]  # matrix is singular, so last index is ~0


def superstage_rates(R, S, superstages, save_time=None):
    """Compute rate for a set of ion superstages. 
    Input and output rates are log-values in arbitrary base.

    Parameters
    ----------
    R : array (time,nZ,space)
        Array containing the effective recombination rates for all ion stages, 
        These are typically combinations of radiative and dielectronic recombination, 
        possibly also of charge exchange recombination.
    S : array (time,nZ,space)
        Array containing the effective ionization rates for all ion stages.
    superstages : list or 1D array
        Indices of charge states of chosen ion that should be included. 
    save_time : list or 1D array of bools
        Indices of the timeslcies which are actually returned by AURORA
    
    Returns
    -------
    superstages : array
        Set of superstages including 0,1 and final stages if these were missing in the input.
    R_rates_super : array (time,nZ-super,space)
        effective recombination rates for superstages
    S_rates_super : array (time,nZ-super,space)
        effective ionization rates for superstages
    fz_upstage : array (space, nZ, time_
        fractional abundances of stages within superstages

    """
    Z_imp = S.shape[1]

    # check input superstages
    if 1 not in superstages:
        print("Warning: 1th superstage was included")
        superstages = np.r_[1, superstages]
    if 0 not in superstages:
        print("Warning: 0th superstage was included")
        superstages = np.r_[0, superstages]
    if np.any(np.diff(superstages) <= 0):
        print("Warning: sorted superstages in increasing order")
        superstages = np.sort(superstages)
    if superstages[-1] > Z_imp:
        raise Exception("The highest superstage must be less than Z_imp = %d" % Z_imp)
    superstages = np.array(superstages)

    # indexing that accounts for no recomb of neutral stage or ionization of full-stripped stage
    R_rates_super = R[:, superstages[1:] - 1]
    S_rates_super = S[:, superstages[1:] - 1]

    if len(S) == 1 or save_time is None:  # time averaged kinetic profiles
        t_slice = slice(None, None)
        nt = 1
    else:  # time resolved kinetic profiles
        t_slice = save_time
        nt = save_time.sum()

    # fractional abundance of supestages used for upstaging.
    fz_upstage = np.ones((R.shape[-1], Z_imp + 1, nt))

    # add fully-stripped charge state
    _superstages = np.r_[superstages, Z_imp + 1]

    for i in range(len(_superstages) - 1):
        if _superstages[i] + 1 != _superstages[i + 1]:
            sind = slice(_superstages[i] - 1, _superstages[i + 1] - 1)

            # calculate fractional abundances within the superstage
            rate_ratio = S[:, sind] / R[:, sind]
            fz = np.cumprod(rate_ratio, axis=1)
            fz /= np.maximum(1e-60, fz.sum(1))[:, None]  # prevents zero division

            # superstage rates
            R_rates_super[:, i - 1] *= np.maximum(fz[:, 0], 1e-60)

            if i < len(_superstages) - 2:  # last superstage cannot ionize further
                S_rates_super[:, i] *= np.maximum(fz[:, -1], 1e-60)

            # fractional abundances inside of each superstage
            fz_upstage[:, _superstages[i] : _superstages[i + 1]] = fz.T[:, :, t_slice]

    return superstages, R_rates_super, S_rates_super, fz_upstage


def get_frac_abundances(
    atom_data,
    ne_cm3,
    Te_eV=None,
    Ti_eV=None,
    n0_by_ne=0.0,
    superstages=[],
    ne_tau=np.inf,
    plot=True,
    ax=None,
    rho=None,
    rho_lbl=None,
):
    r"""Calculate fractional abundances from ionization and recombination equilibrium.
    If n0_by_ne is not 0, radiative recombination and thermal charge exchange are summed.

    This method can work with ne,Te and n0_by_ne arrays of arbitrary dimension, but plotting 
    is only supported in 1D (defaults to flattened arrays).

    Parameters
    ----------
    atom_data : dictionary of atomic ADAS files (only acd, scd are required; ccd is 
        necessary only if include_cx=True)
    ne_cm3 : float or array
        Electron density in units of :math:`cm^{-3}`
    Te_eV : float or array, optional
        Electron temperature in units of eV. If left to None, the Te grid given in the 
        atomic data is used.
    Ti_eV : float or array, optional
        Bulk ion temperature in units of eV. If left to None, Ti is set to be equal to Te
    n0_by_ne: float or array, optional
        Ratio of background neutral hydrogen to electron density. If not 0, CX is considered.
    superstages : list or 1D array
        Indices of charge states of chosen ion that should be included. If left empty, all ion stages
        are included. If only some indices are given, these are modeled as "superstages".
    ne_tau : float, opt
        Value of electron density in :math:`m^{-3}\cdot s` :math:`\times` particle residence time. 
        This is a scalar value that can be used to model the effect of transport on ionization equilibrium. 
        Setting ne_tau=np.inf (default) corresponds to no effect from transport. 
    plot : bool, optional
        Show fractional abundances as a function of ne,Te profiles parameterization.
    ax : matplotlib.pyplot Axes instance
        Axes on which to plot if plot=True. If False, it creates new axes
    rho : list or array, optional
        Vector of radial coordinates on which ne,Te (and possibly n0_by_ne) are given. 
        This is only used for plotting, if given. 
    rho_lbl: str, optional
        Label to be used for rho. If left to None, defaults to a general "x".

    Returns
    -------
    Te : array
        electron temperatures as a function of which the fractional abundances and
        rate coefficients are given.
    fz : array, (space,nZ)
        Fractional abundances across the same grid used by the input ne,Te values. 

    """
    # if input arrays are multi-dimensional, flatten them here and restructure at the end
    _ne = np.ravel(ne_cm3)
    _Te = np.ravel(Te_eV) if Te_eV is not None else None
    _Ti = np.ravel(Ti_eV) if Ti_eV is not None else _Te
    _n0_by_ne = np.ravel(n0_by_ne)
    if superstages is None:
        superstages = []

    include_cx = False if not np.any(n0_by_ne) else True

    Te, Sne, Rne, cxne = get_cs_balance_terms(
        atom_data, _ne, _Te, _Ti, include_cx=include_cx
    )
    Z_imp = Sne.shape[1]

    if include_cx:
        # Get an effective recombination rate by summing radiative & CX recombination rates
        Rne += n0_by_ne[:, None] * cxne

    rate_ratio = np.hstack((np.ones_like(Te)[:, None], Sne / Rne))
    fz_full = np.cumprod(rate_ratio, axis=1)
    fz_full /= fz_full.sum(1)[:, None]

    # Enable use of superstages
    if len(superstages):

        # superstage_rates expects values in shape (time,nZ,space)
        superstages, Rne, Sne, _ = superstage_rates(
            Rne.T[None], Sne.T[None], superstages
        )
        Rne = Rne[0].T
        Sne = Sne[0].T

        rate_ratio = np.hstack((np.ones_like(Te)[:, None], Sne / Rne))
        fz_super = np.cumprod(rate_ratio, axis=1)
        fz_super /= fz_super.sum(1)[:, None]

        # bundled stages can have very high values -- clip here
        Rne = np.clip(Rne, 1e-25, 1)
        Sne = np.clip(Sne, 1e-25, 1)

        _superstages = np.r_[superstages, Z_imp + 1]

    if plot:
        # plot fractional abundances (only 1D)
        if ax is None:
            fig, axx = plt.subplots()
        else:
            axx = ax

        if rho is None:
            x = Te
            axx.set_xlabel("T$_e$ [eV]")
            axx.set_xscale("log")
        else:
            if rho_lbl is None:
                rho_lbl = "x"
            x = rho
            axx.set_xlabel(rho_lbl)

        axx.set_prop_cycle("color", cm.plasma(np.linspace(0, 1, fz_full.shape[1])))

        css = 0
        for cs in range(fz_full.shape[1]):
            l = axx.semilogy(x, fz_full[:, cs], ls="--")
            imax = np.argmax(fz_full[:, cs])
            axx.text(
                np.max([0.1, x[imax]]),
                fz_full[imax, cs],
                cs,
                horizontalalignment="center",
                clip_on=True,
            )

            if len(superstages) and cs in _superstages:
                axx.semilogy(x, fz_super[:, css], c=l[0].get_color(), ls="-")
                imax = np.argmax(fz_super[:, css])

                if _superstages[css] == Z_imp:
                    lbl = r"\{" + f"{_superstages[css]}" + r"\}"
                else:
                    if _superstages[css] != _superstages[css + 1] - 1:
                        lbl = (
                            r"\{"
                            + f"{_superstages[css]},{_superstages[css+1]-1}"
                            + r"\}"
                        )
                    else:
                        lbl = r"\{" + f"{_superstages[css]}" + r"\}"
                axx.text(
                    np.max([0.05, x[imax]]),
                    fz_super[imax, css],
                    lbl,
                    horizontalalignment="center",
                    clip_on=True,
                    backgroundcolor="w",
                )
                css += 1

        axx.grid("on")
        axx.set_ylim(1e-2, 1.5)
        axx.set_xlim(x[0], x[-1])
        plt.tight_layout()

    if np.size(ne_cm3) > 1:
        # re-structure to original array dimensions
        Te = Te.reshape(np.array(ne_cm3).shape)
        fz_full = fz_full.reshape(*np.array(ne_cm3).shape, fz_full.shape[1])
        if len(superstages):
            fz_super = fz_super.reshape(*np.array(ne_cm3).shape, fz_super.shape[1])

    return [Te,] + [
        fz_super if len(superstages) else fz_full,
    ]


def get_cs_balance_terms(
    atom_data, ne_cm3=5e13, Te_eV=None, Ti_eV=None, include_cx=True
):
    """Get S*ne, R*ne and cx*ne rates on the same logTe grid. 
    
    Parameters
    ----------
    atom_data : dictionary of atomic ADAS files (only acd, scd are required; ccd is 
        necessary only if include_cx=True)
    ne_cm3 : float or array
        Electron density in units of :math:`cm^{-3}`
    Te_eV : float or array
        Electron temperature in units of eV. If left to None, the Te grid
        given in the atomic data is used.
    Ti_eV : float or array
        Bulk ion temperature in units of eV, only needed for CX. If left to None, Ti is set equal to Te.
    include_cx : bool
        If True, obtain charge exchange terms as well. 
    
    Returns
    -------
    Te : array (n_Te)
        Te grid on which atomic rates are given
    Sne, Rne (,cxne): arrays (n_ne,n_Te)
        atomic rates for effective ionization, radiative+dielectronic
        recombination (+ charge exchange, if requested). 
        All terms will be in units of :math:`s^{-1}`. 

    Notes
    -----
        The nomenclature with 'ne' at the end of rate names
        indicates that these are rates in units of :math:`m^3\cdot s^{-1}` multiplied by the
        electron density 'ne' (hence, final units of :math:`s^{-1}`).
    """

    if Te_eV is None:
        # find smallest Te grid from all files
        _, logTe1, _ = atom_data["scd"]
        _, logTe2, _ = atom_data["acd"]

        minTe = max(logTe1[0], logTe2[0])
        maxTe = min(logTe1[-1], logTe2[-1])  # avoid extrapolation

        if include_cx:
            _, logTe3, _ = atom_data["ccd"]  # thermal cx recombination
            minTe = max(minTe, logTe3[0])
            maxTe = min(maxTe, logTe3[-1])  # avoid extrapolation

        Te_eV = np.logspace(minTe, maxTe, 200)

    logne = np.log10(ne_cm3)
    logTe = np.log10(Te_eV)

    Sne = interp_atom_prof(atom_data["scd"], logne, logTe, x_multiply=True)
    Rne = interp_atom_prof(atom_data["acd"], logne, logTe, x_multiply=True)
    if include_cx:
        logTi = np.log10(Ti_eV) if Ti_eV is not None else logTe
        x, y, tab = atom_data["ccd"]
        # select appropriate number of charge states
        # this allows use of CCD files from higher-Z ions because of simple CX scaling
        cxne = interp_atom_prof(
            (x, y, tab[: Sne.shape[1]]), logne, logTi, x_multiply=True
        )
    else:
        cxne = None

    return Te_eV, Sne, Rne, cxne


def get_atomic_relax_time(
    atom_data,
    ne_cm3,
    Te_eV=None,
    Ti_eV=None,
    n0_by_ne=0.0,
    superstages=[],
    ne_tau=np.inf,
    plot=True,
    ax=None,
    ls="-",
):
    r"""Obtain the relaxation time of the ionization equilibrium for a given atomic species.

    If n0_by_ne is not 0, thermal charge exchange is added to radiative and dielectronic recombination.
    This function can work with ne,Te and n0_by_ne arrays of arbitrary dimension.
    It uses a matrix SVD approach in order to find the relaxation rates, as opposed to the simpler
    approach of :py:func:`~aurora.get_frac_abundances`, but fractional abundances produced by the two
    methods should be the same.

    Parameters
    ----------
    atom_data : dictionary of atomic ADAS files (only acd, scd are required; ccd is 
        necessary only if n0_by_ne is not 0).
    ne_cm3 : float or array
        Electron density in units of :math:`cm^{-3}`.
    Te_eV : float or array, optional
        Electron temperature in units of eV. If left to None, the Te grid given in the atomic data is used.
    Ti_eV : float or array
        Bulk ion temperature in units of eV, only needed for CX. If left to None, Ti is set equal to Te.
    n0_by_ne: float or array, optional
        Ratio of background neutral hydrogen to electron density. If set to 0, CX is not considered.
    superstages : list or 1D array
        Indices of charge states of chosen ion that should be included. If left empty, all ion stages
        are included. If only some indices are given, these are modeled as "superstages".
    ne_tau : float, opt
        Value of electron density in :math:`m^{-3}\cdot s` :math:`\times` particle residence time. 
        This is a scalar value that can be used to model the effect of transport on ionization equilibrium. 
        Setting ne_tau=np.inf (default) corresponds to no effect from transport. 
    plot : bool, optional
        If True, the atomic relaxation time is plotted as a function of Te. Default is True.
    ax : matplotlib.pyplot Axes instance
        Axes on which to plot if plot=True. If False, new axes are created.
    ls : str, optional
        Line style for plots. Continuous lines are used by default. 

    Returns
    -------
    Te : array
        electron temperatures as a function of which the fractional abundances and
        rate coefficients are given.
    fz : array, (space,nZ)
        Fractional abundances across the same grid used by the input ne,Te values. 
    rate_coeffs : array, (space, nZ)
        Rate coefficients in units of [:math:`s^{-1}`]. 
    """
    # if input arrays are multi-dimensional, flatten them here and restructure at the end
    _ne = np.ravel(ne_cm3)
    _Te = np.ravel(Te_eV) if Te_eV is not None else None
    _Ti = np.ravel(Ti_eV) if Ti_eV is not None else _Te
    _n0_by_ne = np.ravel(n0_by_ne)

    include_cx = False if not np.any(n0_by_ne) else True

    Te, Sne, Rne, cxne = get_cs_balance_terms(
        atom_data, _ne, _Te, include_cx=include_cx
    )

    if include_cx:
        # Get an effective recombination rate by summing radiative & CX recombination rates
        Rne += cxne * _n0_by_ne

    # Enable use of superstages
    if len(superstages):
        _, Rne, Sne, _ = superstage_rates(Rne, Sne, superstages)

    # numerical method that calculates also rate_coeffs
    nion = Rne.shape[1]
    fz = np.zeros((Te.size, nion + 1))
    rate_coeffs = np.zeros(Te.size)

    for it, t in enumerate(Te):
        A = (
            -np.diag(np.r_[Sne[it], 0] + np.r_[0, Rne[it]] + 1e6 / ne_tau)
            + np.diag(Sne[it], -1)
            + np.diag(Rne[it], 1)
        )

        N, rate_coeffs[it] = null_space(A)
        fz[it] = N / np.sum(N)

    if np.size(ne_cm3) > 1:
        # re-structure to original array dimensions
        Te = Te.reshape(np.shape(ne_cm3))
        fz = fz.reshape(*np.shape(ne_cm3), fz.shape[1])
        rate_coeffs = rate_coeffs.reshape(np.shape(ne_cm3))

    if plot:
        # Now plot relaxation times
        if ax is None:
            fig, ax = plt.subplots()

        ax.loglog(Te, 1e3 / rate_coeffs, "b")
        ax.set_xlim(Te[0], Te[-1])
        ax.grid("on")
        ax.set_xlabel("T$_e$ [eV]")
        ax.set_ylabel(r"$\tau_\mathrm{relax}$ [ms]")

    return Te, fz, rate_coeffs


class CartesianGrid:
    """Fast linear interpolation for 1D and 2D vector data on equally spaced grids.
    This offers optimal speed in Python for interpolation of atomic data tables such
    as the ADAS ones.

    Parameters
    ----------
    grids: list of arrays, N=len(grids), N=1 or N=2
        List of 1D arrays with equally spaced grid values for each dimension
    values: N+1 dimensional array of values used for interpolation
        Values to interpolate. The first dimension typically refers to different ion stages, for which 
        data is provided on the input grids. 
        Other dimensions refer to values on the density and temperature grids.
    """

    def __init__(self, grids, values):

        values = np.ascontiguousarray(np.moveaxis(values, 0, -1))
        self.N = values.shape[:-1]

        if len(self.N) > 2:
            raise OMFITexception("Only 1 and 2 dimensional interpolation is supported")

        for g, s in zip(grids, self.N):
            if len(g) != s:
                raise OMFITexception("wrong size of values array")

        self.eq_spaced_grid = np.all(
            [np.std(np.diff(g)) / np.std(g) < 0.01 for g in grids]
        )
        if self.eq_spaced_grid:
            self.offsets = [g[0] for g in grids]
            self.scales = [(g[-1] - g[0]) / (n - 1) for g, n in zip(grids, self.N)]
        else:
            self.grids = grids

        A = []
        if len(self.N) == 1:
            A.append(values[:-1])
            A.append(values[1:] - A[0])

        if len(self.N) == 2:
            A.append(values[:-1, :-1])
            A.append(values[1:, :-1] - A[0])
            A.append(values[:-1, 1:] - A[0])
            A.append(values[1:, 1:] - A[2] - A[1] - A[0])

        self.A = A

    def __call__(self, *coords):
        """Evaluate the interpolation at the input `coords` values.
        
        This offers optimally-fast linear interpolation for 1D and  2D dimensional vector data 
        on a equally spaced grids
        
        Parameters
        ----------
        coords:  list of arrays
            List of 1D arrays for the N coordines (N=1 or N=2). These arrays must be of the same shape.
 
        """

        if self.eq_spaced_grid:
            coords = np.array(coords).T
            coords -= self.offsets
            coords /= self.scales
            coords = coords.T
        else:
            # for non-equally spaced grids must be used linear interpolation to map inputs to equally spaced indexes
            coords = [
                np.interp(c, g, np.arange(len(g))) for c, g in zip(coords, self.grids)
            ]

        # clip dimension - it will extrapolation by a nearest value
        for coord, n in zip(coords, self.N):
            np.clip(coord, 0, n - 1.00001, coord)

        #  en.wikipedia.org/wiki/Bilinear_interpolation#Unit_square

        # get indicies x and weights dx
        x = np.int16(coords)  # fast floor(x) function
        coords -= x  # frac(x)
        dx = coords[..., None]

        # prepare coefficients
        if len(self.N) == 1:  # linear interpolation
            # inplace evaluate linear  interpolation
            inter_out = self.A[1][x[0]]
            inter_out *= dx[0]
            inter_out += self.A[0][x[0]]

        else:  # bilinear interpolation
            # inplace evaluate linear  interpolation
            inter_out = self.A[1][x[0], x[1]]
            inter_out *= dx[0]
            inter_out += self.A[0][x[0], x[1]]

            _tmp = self.A[3][x[0], x[1]]
            _tmp *= dx[0]
            _tmp += self.A[2][x[0], x[1]]
            _tmp *= dx[1]
            inter_out += _tmp

        return np.moveaxis(inter_out, -1, 0)


def interp_atom_prof(atom_table, xprof, yprof, log_val=False, x_multiply=True):
    r""" Fast interpolate atomic data in atom_table onto the xprof and yprof profiles.
    This function assume that xprof, yprof, x, y, table are all base-10 logarithms,
    and xprof, yprof are equally spaced.

    Parameters
    ----------
    atom_table : list
        List with x,y, table = atom_table, containing atomic data from one of the ADAS files. 
    xprof : array (nt,nr)
        Spatio-temporal profiles of the first coordinate of the ADAS file table (usually 
        electron density in :math:`cm^{-3}`)
    yprof : array (nt,nr)
        Spatio-temporal profiles of the second coordinate of the ADAS file table (usually 
        electron temperature in :math:`eV`)
    log_val : bool
        If True, return natural logarithm of the data
    x_multiply : bool
        If True, multiply output by :math:`10^{xprof}`. 

    Returns
    -------
    interp_vals : array (nt,nion,nr)
        Interpolated atomic data on time,charge state and spatial grid that correspond to the 
        ion of interest and the spatiotemporal grids of xprof and yprof.

    Notes
    -------
    This function uses `np.log10` and exponential operations to optimize speed, since it has
    been observed that base-e operations are faster than base-10 operations in numpy. 
    """
    x, y, table = atom_table

    if x_multiply:  # multiplying of logarithms is just adding
        table = table + x  # don't modify original table, create copy

    if (abs(table - table[..., [0]]) < 0.05).all() or xprof is None:
        # 1D interpolation if independent of the last dimension - like SXR radiation data

        reg_interp = CartesianGrid((y,), table[:, :, 0] * np.log(10))
        interp_vals = reg_interp(yprof)

    else:  # 2D interpolation
        # broadcast both variables to the same shape
        xprof, yprof = np.broadcast_arrays(xprof, yprof)

        # perform fast linear interpolation
        reg_interp = CartesianGrid((x, y), table.swapaxes(1, 2) * np.log(10))

        interp_vals = reg_interp(xprof, yprof)

    # reshape to shape(nt,nion,nr)
    interp_vals = interp_vals.swapaxes(0, 1)

    if not log_val:
        # return actual value, not logarithm
        np.exp(interp_vals, out=interp_vals)

    return interp_vals


def gff_mean(Z, Te):
    """
    Total free-free gaunt factor yielding the total radiated bremsstrahlung power
    when multiplying with the result for gff=1.
    Data originally from Karzas & Latter, extracted from STRAHL's atomic_data.f.
    """
    from scipy.constants import e, h, c, Rydberg

    thirteenpointsix = h * c * Rydberg / e

    log_gamma2_grid = [
        -3.000,
        -2.833,
        -2.667,
        -2.500,
        -2.333,
        -2.167,
        -2.000,
        -1.833,
        -1.667,
        -1.500,
        -1.333,
        -1.167,
        -1.000,
        -0.833,
        -0.667,
        -0.500,
        -0.333,
        -0.167,
        0.000,
        0.167,
        0.333,
        0.500,
        0.667,
        0.833,
        1.000,
    ]

    gff_mean_grid = [
        1.140,
        1.149,
        1.159,
        1.170,
        1.181,
        1.193,
        1.210,
        1.233,
        1.261,
        1.290,
        1.318,
        1.344,
        1.370,
        1.394,
        1.416,
        1.434,
        1.445,
        1.448,
        1.440,
        1.418,
        1.389,
        1.360,
        1.336,
        1.317,
        1.300,
    ]

    # set min Te here to 10 eV, because the grid above does not extend to lower temperatures
    Te = np.maximum(Te, 10.0)

    log_gamma2 = np.log10(Z ** 2 * thirteenpointsix / Te)

    # dangerous/inaccurate extrapolation...
    return np.interp(log_gamma2, log_gamma2_grid, gff_mean_grid)


def impurity_brems(nz, ne, Te):
    """Approximate bremsstrahlung in units of :math:`mW/nm/sr/m^3 \cdot cm^3`, or
    equivalently :math:`W/m^3` for full spherical emission. 

    Note that this may not be very useful, since this contribution is already included 
    in the continuum radiation component in ADAS files. The bremsstrahlung estimate in 
    ADAS continuum radiation files is more accurate than the one give by this function, 
    which uses a simpler interpolation of the Gaunt factor with weak ne-dependence. 
    Use with care!

    Parameters
    ----------
    nz : array (time,nZ,space)
        Densities for each charge state [:math:`cm^{-3}`]
    ne : array (time,space)
        Electron density [:math:`cm^{-3}`]
    Te : array (time,space)
        Electron temperature [:math:`cm^{-3}`]

    Returns
    -------
    brems : array (time,nZ,space)
        Bremsstrahlung for each charge state 
    """
    # neutral stage doesn't produce brems
    Z_imp = nz.shape[1] - 1
    Z = np.arange(Z_imp)[None, :, None] + 1

    gff = gff_mean(Z, Te[:, None])
    brems = Z ** 2 * nz[:, 1:] * gff * (1.69e-32 * np.sqrt(Te) * ne)[:, None]

    return brems


def plot_norm_ion_freq(
    S_z,
    q_prof,
    R_prof,
    imp_A,
    Ti_prof,
    nz_profs=None,
    rhop=None,
    plot=True,
    eps_prof=None,
):
    r"""Compare effective ionization rate for each charge state with the characteristic
    transit time that passing and trapped impurity ions take to travel a parallel distance 
    :math:`L = q R`, defining

    .. math::

        \nu_{ion}^* \equiv \nu_{ion} \tau_t = \nu_{ion} \frac{q R}{v_{th}} = \frac{\sum_z n_z \nu_z^{ion}}{\sum_z n_z} q R \sqrt{\frac{m_{imp}}{2 k_B T_i}}

    following Ref.[1]_. If the normalized ionization rate (:math:`\nu_{ion}^*`) is less than 1, 
    then flux surface averaging of background asymmetries (e.g. from edge or beam neutrals) may 
    be taken as a good approximation of reality; in this case, 1.5D simulations of impurity transport
    are expected to be valid. If, on the other hand, :math:`\nu_{ion}^*>1` then local effects 
    may be too important to ignore. 

    Parameters
    ----------
    S_z : array (r,cs) [:math:`s^{-1}`]
         Effective ionization rates for each charge state as a function of radius. 
         Note that, for convenience within aurora, cs includes the neutral stage.
    q_prof : array (r,)
         Radial profile of safety factor
    R_prof : array (r,) or float [m]
         Radial profile of major radius, either given as an average of HFS and LFS, or also
         simply as a scalar (major radius on axis)
    imp_A : float [amu]
         Atomic mass number, i.e. number of protons + neutrons (e.g. 2 for D)
    Ti_prof : array (r,)
         Radial profile of ion temperature [:math:`eV`]
    nz_profs : array (r,cs), optional
         Radial profile for each charge state. If provided, calculate average normalized 
         ionization rate over all charge states.
    rhop : array (r,), optional
         Sqrt of poloidal flux radial grid. This is used only for (optional) plotting. 
    plot : bool, optional
         If True, plot results.
    eps_prof : array (r,), optional
         Radial profile of inverse aspect ratio, i.e. r/R, only used if plotting is requested.  

    Returns
    -------
    nu_ioniz_star : array (r,cs) or (r,)
         Normalized ionization rate. If nz_profs is given as an input, this is an average over
         all charge state; otherwise, it is given for each charge state.

    References
    ----------
    .. [1] R.Dux et al. Nucl. Fusion 60 (2020) 126039

    """
    nu = np.zeros(
        (S_z.shape[0], S_z.shape[1] - 1)
    )  # exclude neutral states, which have no parallel transport
    for cs in np.arange(nu.shape[1]):
        nu[:, cs] = (
            S_z[:, cs + 1]
            * q_prof
            * R_prof
            * np.sqrt((imp_A * constants.m_p) / (2 * Ti_prof * constants.e))
        )

    if nz_profs is not None:
        # calculate average nu_ioniz_star
        nu_ioniz_star = np.sum(nz_profs[:, 1:] * nu, axis=1) / np.sum(
            nz_profs[:, 1:], axis=1
        )
    else:
        # return normalized ionization rate for each charge state
        nu_ioniz_star = nu

    if plot:
        if rhop is None:
            rhop = np.arange(nu.shape[0])

        fig, ax = plt.subplots()
        if nu_ioniz_star.ndim == 1:
            ax.semilogy(rhop, nu_ioniz_star, label=r"$\nu_{ion}^*$")
        else:
            for cs in np.arange(nu.shape[1]):
                ax.semilogy(rhop, nu_ioniz_star[:, cs], label=f"q={cs+1}")
            ax.set_ylabel(r"$\nu_{ion}^*$")

        ax.set_xlabel(r"$\rho_p$")

        if eps_prof is not None:
            ax.semilogy(rhop, np.sqrt(eps_prof), label=r"$\sqrt{\epsilon}$")

        ax.legend().set_draggable(True)
        ax.set_xlim([0, 1])


def read_adf12(filename, block, Ebeam, ne_cm3, Ti_eV, zeff):
    """Read charge exchange effective emission coefficients from ADAS ADF12 files.

    See https://open.adas.ac.uk/adf12 for details and to download files.

    Parameters
    ----------
    filename : str
        adf12 file name/path
    block : int
        Source block selected
    Ebeam : float
        Energy of the neutral beam population of interest, in units of :math:`eV/amu`.
    ne_cm3 : float or 1D array
        Electron densities at which to evaluate coefficients, in units of :math:`cm^{-3}`.
    Ti_eV : float or 1D array
        Bulk ion temperature at which to evaluate coefficients, in units of :math:`eV`.
    Zeff : float or 1D array
        Effective background charge.

    Returns
    -------
    float or 1D array :
        Interpolated coefficients, units of :math:`cm^3/s`.

    """
    with open(filename, "r") as f:
        nlines = int(f.readline())

        for iline in range(block):
            cer_line = {}
            params = []
            first_line = "0"
            while not first_line[0].isalpha():
                first_line = f.readline()

            cer_line["header"] = first_line
            cer_line["qefref"] = np.float(f.readline()[:63].replace("D", "e"))
            cer_line["parmref"] = np.float_(f.readline()[:63].replace("D", "e").split())
            cer_line["nparmsc"] = np.int_(f.readline()[:63].split())

            for ipar, npar in enumerate(cer_line["nparmsc"]):
                for q in range(2):
                    data = []
                    while npar > len(data):
                        line = f.readline()
                        if len(line) > 63:
                            name = line[63:].strip().lower()
                            cer_line[name] = []
                            if q == 0:
                                params.append(name)

                        values = np.float_(line[:63].replace("D", "E").split())
                        values = values[values > 0]
                        if not len(values):
                            continue
                        data += values.tolist()
                    cer_line[name] = data

    # interpolate in logspace
    lqefref = np.log(cer_line["qefref"])

    lnq = np.zeros(np.broadcast(Ebeam, ne_cm3, Ti_eV, zeff).shape)
    lnq += lqefref * (1 - 4)
    lnq += np.interp(np.log(Ti_eV), np.log(cer_line["tiev"]), np.log(cer_line["qtiev"]))
    lnq += np.interp(
        np.log(ne_cm3), np.log(cer_line["densi"]), np.log(cer_line["qdensi"])
    )
    lnq += np.interp(np.log(Ebeam), np.log(cer_line["ener"]), np.log(cer_line["qener"]))
    lnq += np.interp(np.log(zeff), np.log(cer_line["zeff"]), np.log(cer_line["qzeff"]))

    return np.exp(lnq)


def read_adf21(filename, Ebeam, ne_cm3, Te_eV):
    """Read ADAS ADF21 or ADF22 files.

    ADF21 files contain effective beam stopping/excitation coefficients.
    ADF22 contain effective beam emission/population coefficients.

    See https://open.adas.ac.uk/adf21 and https://open.adas.ac.uk/adf22 for details
    and to download files.

    Parameters
    ----------
    filename : str
        adf21 or adf22 file name/path
    Ebeam : float
        Energy of the neutral beam, in units of :math:`eV/amu`.
    ne_cm3 : float or 1D array
        Electron densities at which to evaluate coefficients.
    Te_eV : float or 1D array
        Electron temperature at which to evaluate coefficients.

    Returns
    -------
    float or 1D array :
        Interpolated coefficients. For ADF21 files, these have units of :math:`cm^3/s`for ADF21 files.
        For ADF22, they correspond to n=2 fractional abundances.
    """

    with open(filename, "r") as f:
        line = f.readline()
        ref = float(line.split()[1].split("=")[1])
        f.readline()
        line = f.readline()
        nE, nne, Teref = line.split()
        nE, nne = int(nE), int(nne)
        Teref = float(Teref.split("=")[1])
        f.readline()

        E = []
        while len(E) < nE:
            line = f.readline()
            E.extend([float(f) for f in line.split()])
        E = np.array(E)

        ne = []
        while len(ne) < nne:
            line = f.readline()
            ne.extend([float(f) for f in line.split()])
        ne = np.array(ne)
        f.readline()

        Q2 = []
        while len(Q2) < nne * nE:
            line = f.readline()
            Q2.extend([float(f) for f in line.split()])
        Q2 = np.reshape(Q2, (nne, nE))

        f.readline()
        line = f.readline()
        nTe, Eref, Neref = line.split()
        nTe, Eref, Neref = (
            int(nTe),
            float(Eref.split("=")[1]),
            float(Neref.split("=")[1]),
        )

        f.readline()

        Te = []
        while len(Te) < nTe:
            line = f.readline()
            Te.extend([float(f) for f in line.split()])
        Te = np.array(Te)

        f.readline()

        Q1 = []
        while len(Q1) < nTe:
            line = f.readline()
            Q1.extend([float(f) for f in line.split()])
        Q1 = np.array(Q1)

    # clip data in available range to avoid extrapolation
    Ebeam = np.clip(Ebeam, *E[[0, -1]])
    ne_cm3 = np.clip(ne_cm3, *ne[[0, -1]])
    Te_eV = np.clip(Te_eV, *Te[[0, -1]])

    lref = np.log(ref)

    # interpolate on the requested values
    RectInt1 = interp.interp1d(
        np.log(Te), np.log(Q1) - lref, assume_sorted=True, kind="quadratic"
    )
    RectInt2 = interp.RectBivariateSpline(
        np.log(ne), np.log(E), np.log(Q2) - lref, kx=2, ky=2
    )

    adf = RectInt1(np.log(Te_eV)) + RectInt2.ev(np.log(ne_cm3), np.log(Ebeam))
    return np.exp(adf + lref)


def get_natural_partition(ion, plot=True):
    """Identify natural partition of charge states by plotting the variation of ionization energy
    as a function of charge for a given ion, using the ADAS metric :math:`2 (I_{z+1}-I_z)/(I_{z+1}+I_z)`.
    
    Parameters
    ----------
    ion : str
        Atomic symbol of species of interest.
    plot : bool
        If True, plot the variation of ionization energy.

    Returns
    -------
    q : 1D array
        Metric to identify natural partition.
    
    Notes
    -----
    A ColRadPy installation must be available for this function to work.
    """

    try:
        # temporarily import this here, until ColRadPy dependency can be set up properly
        import colradpy
    except ImportError:
        raise ValueError(
            "Could not import colradpy. Install this from the Github repo!"
        )

    # find location of files containing NIST data
    colradpy_dist = os.sep.join(colradpy.__file__.split(os.sep)[:-2])
    loc = colradpy_dist + os.sep + "atomic" + os.sep + "nist_energies"

    # find energies for each charge state of interest
    _E_eV = []
    cs = []
    for filename in os.listdir(loc):
        if not filename.startswith("#") and not filename.endswith("~"):
            _ion = filename.split("_")[0]

            if _ion != ion.lower():
                continue
            charge = int(filename.split("_")[1])

            # read energy from file
            with open(loc + os.sep + filename, "r") as f:
                cont = f.readlines()

            # last value is energy in cm^-1
            E_cm_val = float(cont[-1].split(",")[-1].strip("/n"))
            E_eV_val = constants.h * constants.c / constants.e * (E_cm_val * 100.0)

            cs.append(charge)
            _E_eV.append(E_eV_val)

    idx = np.argsort(cs)
    E_eV = np.array(_E_eV)[idx]

    # compute ADAS natural partitioning metric
    q = []
    for i in np.arange(1, len(E_eV) - 1):
        q.append(2 * (E_eV[i + 1] - E_eV[i]) / (E_eV[i + 1] + E_eV[i]))

    if plot:
        fig, ax = plt.subplots()

        ax.plot(np.arange(len(q)), q)
        ax.set_xlabel("Z")
        ax.set_title(r"$2\times (I_{z+1}-I_z)/(I_{z+1}+I_z)$")

        # take running mean over 7 adjacent charge states as indicated by Foster's thesis
        q_mean = np.convolve(q, np.ones(7) / 7, mode="same")
        plt.plot(q_mean)

    return q
